import logging

import tensorflow_probability.substrates.jax as tfp
from jax import random, numpy as jnp, vmap

from jaxns.prior import PriorModelGen, Prior, parse_prior, compute_log_likelihood, InvalidDistribution, \
    InvalidPriorName, prepare_input, distribution_chain
from jaxns.special_priors import Bernoulli, Categorical, Poisson, Beta, ForcedIdentifiability
from jaxns.types import float_type

logger = logging.getLogger('jaxns')
tfpd = tfp.distributions


def test_single_prior():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2), name='x')
        y = yield Prior(tfpd.Normal(loc=2, scale=x), name='x')
        z = x + y
        return z

    try:
        parse_prior(prior_model)
        assert False
    except InvalidPriorName:
        assert True

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2), name='x')
        y = yield Prior(tfpd.Normal(loc=2, scale=x), name='y')
        z = x + y
        return z

    U, X = parse_prior(prior_model)
    U_ndims = U.shape[0]
    assert isinstance(X, dict)
    assert len(X) == 2
    assert U_ndims == 2

    U = random.uniform(random.PRNGKey(42), shape=(U_ndims,), dtype=float_type)

    def log_likelihood(Z):
        return Z - jnp.sum(Z ** 2)

    log_L = compute_log_likelihood(U, prior_model, log_likelihood=log_likelihood)
    assert not jnp.isnan(log_L)


def test_no_quantile_prior():
    def prior_model() -> PriorModelGen:
        z = yield Prior(tfpd.VonMises(loc=0., concentration=1.))
        return z

    try:
        parse_prior(prior_model)
        assert False
    except InvalidDistribution:
        assert True


def test_prior_model_basic():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z, z ** 2

    U, X = parse_prior(prior_model)
    U_ndims = U.shape[0]
    assert isinstance(X, dict)
    assert len(X) == 0
    assert U_ndims == 2

    U = random.uniform(random.PRNGKey(42), shape=(U_ndims,), dtype=float_type)
    output = prepare_input(U, prior_model)
    assert isinstance(output, tuple)
    assert len(output) == 2
    assert jnp.allclose(output[1], output[0] ** 2)

    def log_likelihood(Z, Z2):
        return Z - jnp.sum(Z2)

    log_L = compute_log_likelihood(U, prior_model, log_likelihood=log_likelihood)
    assert not jnp.isnan(log_L)


def test_distribution_chain():
    d = tfpd.MultivariateNormalTriL(loc=jnp.zeros(5), scale_tril=jnp.eye(5))
    chain = distribution_chain(d)
    assert len(chain) == 2
    assert isinstance(chain[0], tfpd.Sample)
    assert isinstance(chain[1], tfpd.MultivariateNormalTriL)

    chain = distribution_chain(tfpd.Normal(loc=jnp.zeros(5), scale=jnp.ones(5)))
    assert len(chain) == 1
    assert isinstance(chain[0], tfpd.Normal)


def test_priors():
    d = Prior(jnp.zeros(5))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == ()
    assert d.shape == (5,)

    d = Prior(tfpd.Uniform(low=jnp.zeros(5), high=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Normal(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Laplace(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Cauchy(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.StudentT(df=1.5, loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Beta(concentration0=jnp.ones(5), concentration1=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.HalfNormal(scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.HalfCauchy(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Gamma(concentration=jnp.ones(5), rate=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Gumbel(loc=jnp.ones(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == ()
    assert d.shape == (5,)

    d = Prior(tfpd.MultivariateNormalTriL(loc=jnp.zeros(5), scale_tril=jnp.eye(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)


def test_special_priors():
    d = Bernoulli(probs=jnp.ones(5), name='x')
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    probs = jnp.asarray([1, 2, 3, 2, 1], float_type)
    d = Categorical(parametrisation='gumbel_max', probs=probs, name='x')
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == ()
    x = vmap(lambda key: d.forward(random.uniform(key, shape=d.base_shape)))(random.split(random.PRNGKey(42), 10000))
    assert jnp.all(x >= 0)
    assert jnp.all(x < 5)
    assert jnp.any(x == 0)
    assert jnp.any(x == 4)
    count = jnp.bincount(x)
    count /= jnp.sum(count)
    assert jnp.allclose(count, probs / jnp.sum(probs), atol=1e-2)

    probs = jnp.asarray([1, 2, 3, 2, 1], float_type)
    d = Categorical(parametrisation='cdf', probs=probs, name='x')
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == ()
    assert d.shape == ()
    u_array = jnp.linspace(0., 1., 10000)
    x = vmap(lambda u: d.forward(u))(u_array)
    assert jnp.all(x >= 0)
    assert jnp.all(x < 5)
    assert jnp.any(x == 0)
    assert jnp.any(x == 4)
    count = jnp.bincount(x)
    count /= jnp.sum(count)
    assert jnp.allclose(count, probs / jnp.sum(probs), atol=1e-2)

    d = Poisson(rate=jnp.ones(5), name='x')
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Beta(concentration0=jnp.ones(5), concentration1=jnp.ones(5), name='x')
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = ForcedIdentifiability(n=10, low=jnp.zeros(5), high=jnp.ones(5), name='x')
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (10, 5)
    assert d.shape == (10, 5)

    u_input = vmap(lambda key: random.uniform(key, shape=d.base_shape))(random.split(random.PRNGKey(42), 1))
    x = vmap(lambda u: d.forward(u))(u_input)
    u = vmap(lambda x: d.inverse(x))(x)
    assert jnp.allclose(u, u_input)



import logging
from typing import TypeVar, NamedTuple, Tuple, Optional

from etils.array_types import PRNGKey, FloatArray, BoolArray
from jax import numpy as jnp, random, tree_map
from jax._src.lax.control_flow import while_loop

from jaxns.model import Model
from jaxns.static_nested_sampler import MarkovSampler, PreProcessType, SeedPoint
from jaxns.types import Sample, NestedSamplerState
from jaxns.types import float_type, int_type

logger = logging.getLogger('jaxns')

__all__ = ['UniDimSliceSampler', 'MultiDimSliceSampler']

T = TypeVar('T')


class UniDimProposalState(NamedTuple):
    key: jnp.ndarray  # PRNG key
    process_step: jnp.ndarray  # what stage of the sampling each independent point is at
    proposal_count: jnp.ndarray  # how many successful proposals have happened for each independent point
    num_likelihood_evaluations: jnp.ndarray  # how many  likelihood evaluations have been made for each independent point
    point_U0: jnp.ndarray  # the origin of current slice of each independent point
    log_L0: jnp.ndarray  # the value of log-likelihood at each indpendent point origin
    direction: jnp.ndarray  # the direction of slicing
    left: jnp.ndarray  # the left bound of slice
    right: jnp.ndarray  # the right bound of slice
    point_U: jnp.ndarray  # the point up for likelihood computation
    t: jnp.ndarray  # the parameter measuring between left and right bounds.
    log_L_constraint: jnp.ndarray  # the constraint to sample within


class UniDimSliceSampler(MarkovSampler):
    """
    Slice sampler for a single dimension. Produces correlated (non-i.i.d.) samples.
    """

    def __init__(self, model: Model, num_slices: int, midpoint_shrink: bool, perfect: bool,
                 efficiency_threshold: Optional[float] = None):
        """
        Unidimensional slice sampler.

        Args:
            model: Model
            num_slices: number of slices between acceptance, in units of 1, unlike other software which does it in units of prior dimension.
            midpoint_shrink: if true then contract to the midpoint of interval on rejection. Otherwise, contract to
                rejection point.
            perfect: if true then perform exponential shrinkage from maximal bounds, requiring no step-out procedure.
                Otherwise, uses a doubling procedure (exponentially finding bracket).
                Note: Perfect is a misnomer, as perfection also depends on the number of slices between acceptance.
        """
        super().__init__(model=model, efficiency_threshold=efficiency_threshold)
        if num_slices < 1:
            raise ValueError(f"num_slices should be > 0, got {num_slices}.")
        self.num_slices = num_slices
        self.midpoint_shrink = midpoint_shrink
        self.perfect = perfect

    def preprocess(self, state: NestedSamplerState) -> PreProcessType:
        if self.perfect: # nothing needed
            return ()
        # else: # step out with doubling
        #     return multi_ellipsoidal_params()

    def _sample_direction(self, n_key: PRNGKey, ndim: int) -> FloatArray:
        """
        Choose a direction randomly from S^(D-1).

        Args:
            n_key: PRNG key
            ndim: int, number of dimentions

        Returns:
            direction: [D] direction from S^(D-1)
        """
        if ndim == 1:
            return jnp.ones(())
        direction = random.normal(n_key, shape=(ndim,), dtype=float_type)
        direction /= jnp.linalg.norm(direction)
        return direction

    def _slice_bounds(self, point_U0: FloatArray, direction: FloatArray) -> Tuple[FloatArray, FloatArray]:
        """
        Compute the slice bounds, t, where point_U0 + direction * t intersects uit cube boundary.

        Args:
            point_U0: [D]
            direction: [D]

        Returns:
            left_bound: left most point (<= 0).
            right_bound: right most point (>= 0).
        """
        t1 = (1. - point_U0) / direction
        t1_right = jnp.min(jnp.where(t1 >= 0., t1, jnp.inf))
        t1_left = jnp.max(jnp.where(t1 <= 0., t1, -jnp.inf))
        t0 = -point_U0 / direction
        t0_right = jnp.min(jnp.where(t0 >= 0., t0, jnp.inf))
        t0_left = jnp.max(jnp.where(t0 <= 0., t0, -jnp.inf))
        right_bound = jnp.minimum(t0_right, t1_right)
        left_bound = jnp.maximum(t0_left, t1_left)
        return left_bound, right_bound

    def _pick_point_in_interval(self, t_key: PRNGKey, point_U0: FloatArray, direction: FloatArray, left: FloatArray,
                                right: FloatArray) -> Tuple[FloatArray, FloatArray]:
        """
        Select a point along slice in [point_U0 + direction * left, point_U0 + direction * right]

        Args:
            t_key: PRNG key
            point_U0: [D]
            direction: [D]
            left: left most point (<= 0).
            right: right most point (>= 0).

        Returns:
            point_U: [D]
            t: selection point between [left, right]
        """
        t = random.uniform(t_key, minval=left, maxval=right, dtype=float_type)
        point_U = point_U0 + t * direction
        # close_to_zero = (left >= -10*jnp.finfo(left.dtype).eps) & (right <= 10*jnp.finfo(right.dtype).eps)
        # point_U = jnp.where(close_to_zero, point_U0, point_U)
        # t = jnp.where(close_to_zero, jnp.zeros_like(t), t)
        return point_U, t

    def _change_direction(self, from_proposal_state: UniDimProposalState,
                          log_L_proposal: jnp.ndarray) -> UniDimProposalState:
        """
        Successful proposal, but not enough proposals to de-correlate.
        Pick a new direction and propose from current point.
        """
        proposal_key, n_key, t_key = random.split(from_proposal_state.key, 3)
        point_U0 = from_proposal_state.point_U
        log_L0 = log_L_proposal
        direction = self._sample_direction(n_key, point_U0.size)
        # project out the previous direction to sample in orthogonal slice
        if point_U0.size > 1:
            _direction = direction
            direction = direction - direction * (direction @ from_proposal_state.direction)
            direction = jnp.where(jnp.all(direction == jnp.zeros_like(direction)), _direction, direction)
            direction /= jnp.linalg.norm(direction)

        (left, right) = self._slice_bounds(point_U0, direction)
        point_U, t = self._pick_point_in_interval(t_key, point_U0, direction, left, right)

        return from_proposal_state._replace(key=proposal_key,
                                            process_step=jnp.full((), 3, int_type),
                                            point_U0=point_U0,
                                            log_L0=log_L0,
                                            direction=direction,
                                            left=left,
                                            right=right,
                                            point_U=point_U,
                                            t=t)

    def _shrink_interval(self, from_proposal_state: UniDimProposalState, log_L_proposal: jnp.ndarray,
                         log_L_contour: jnp.ndarray,
                         midpoint_shrink: bool, destructive_shrink: bool) -> UniDimProposalState:
        """
        Not successful proposal, so shrink, optionally apply exponential shrinkage.
        """
        # witout exponential shrinkage, we shrink to failed proposal point, which is 100% correct.
        left = jnp.where(from_proposal_state.t < 0., from_proposal_state.t, from_proposal_state.left)
        right = jnp.where(from_proposal_state.t > 0., from_proposal_state.t, from_proposal_state.right)
        key, t_key, midpoint_key = random.split(from_proposal_state.key, 3)
        if destructive_shrink:
            # shrink to constraint using linear approximation
            # Line logic:
            # logL(t) = m * t + b
            # logL(0) = b
            # (logL(t_R) - logL(0))/t_R
            # logL(t_R*alpha) = (logL(t_R) - logL(0))*alpha + logL(0)
            # alpha = (logL_contour - logL(0))/(logL(t_R) - logL(0))
            alpha = (log_L_contour - from_proposal_state.log_L0) / (log_L_proposal - from_proposal_state.log_L0)
            left = jnp.where((from_proposal_state.t < 0.), alpha * left, left)
            right = jnp.where((from_proposal_state.t > 0.), alpha * right, right)
        if midpoint_shrink:
            # we take two points along lines from origin:
            #  - alpha: one from a satisfying point (t=0) to non-satisfying proposal,
            #  - beta: and, one from non-satisfying proposal to the constraint.
            # We shrink to alpha point if beta point is above alpha point.
            # Intuitively, the tangent from constraint is more accurate than proposal and should be a supremum of heights
            # in reasonable cases.
            # An extension is to make alpha shrink to constraint line, which would shrink very fast, but introduce
            # auto-correlation which must be later refined away.
            # Line logic:
            # logL(t) = m * t + b
            # logL(0) = b
            # (logL(t_R) - logL(0))/t_R
            # logL(t_R*alpha) = (logL(t_R) - logL(0))*alpha + logL(0)
            alpha_key, beta_key = random.split(midpoint_key, 2)
            alpha = random.uniform(alpha_key)
            beta = random.uniform(beta_key)
            logL_alpha = from_proposal_state.log_L0 + alpha * (log_L_proposal - from_proposal_state.log_L0)
            logL_beta = log_L_proposal + beta * (log_L_contour - log_L_proposal)
            do_mid_point_shrink = logL_alpha < logL_beta
            left = jnp.where((from_proposal_state.t < 0.) & do_mid_point_shrink, alpha * left, left)
            right = jnp.where((from_proposal_state.t > 0.) & do_mid_point_shrink, alpha * right, right)
        point_U, t = self._pick_point_in_interval(t_key, from_proposal_state.point_U0, from_proposal_state.direction,
                                                  left, right)
        return from_proposal_state._replace(key=key,
                                            process_step=jnp.full((), 3, int_type),
                                            left=left,
                                            right=right,
                                            point_U=point_U,
                                            t=t)

    def get_sample_from_seed(self, key: PRNGKey, seed_point: SeedPoint, log_L_constraint: FloatArray,
                             preprocess_data: PreProcessType) -> Sample:
        slice_sampler_key, select_key, proposal_key, n_key, t_key = random.split(key, 5)

        direction = self._sample_direction(n_key, seed_point.U0.size)
        num_likelihood_evaluations = jnp.full((), 0, int_type)
        (left, right) = self._slice_bounds(seed_point.U0, direction)
        point_U, t = self._pick_point_in_interval(t_key, seed_point.U0, direction, left, right)
        init_proposal_state = UniDimProposalState(key=proposal_key,
                                                  process_step=jnp.full((), 3, int_type),
                                                  proposal_count=jnp.zeros((), int_type),
                                                  num_likelihood_evaluations=num_likelihood_evaluations,
                                                  point_U0=seed_point.U0,
                                                  log_L0=seed_point.log_L0,
                                                  direction=direction,
                                                  left=left,
                                                  right=right,
                                                  point_U=point_U,
                                                  t=t,
                                                  log_L_constraint=log_L_constraint)

        CarryType = Tuple[UniDimProposalState, FloatArray]

        def slice_sampler_body(body_state: CarryType) -> CarryType:
            """
            Iteratively samples num_parallel_samplers new points using a sampler, then replaces that number of points in
            the reservoir.
            """
            (proposal_state, _) = body_state
            log_L_point_U = self.model.forward(proposal_state.point_U)
            num_likelihood_evaluations = proposal_state.num_likelihood_evaluations + jnp.ones_like(
                proposal_state.num_likelihood_evaluations)
            # assumes that log_L0 > log_L_constraint
            good_proposal = jnp.greater(log_L_point_U, proposal_state.log_L_constraint)

            # instability can occur if super close
            potential_instability = jnp.abs(log_L_point_U - proposal_state.log_L_constraint) < jnp.finfo(float_type).eps

            threshold_log_L_constraint = proposal_state.log_L_constraint - jnp.finfo(float_type).eps
            safe_log_L_constraint = jnp.where(potential_instability,
                                              threshold_log_L_constraint,
                                              proposal_state.log_L_constraint)
            proposal_state = proposal_state._replace(log_L_constraint=safe_log_L_constraint)
            good_proposal = good_proposal | potential_instability

            proposal_count = jnp.where(good_proposal,
                                       proposal_state.proposal_count + jnp.ones_like(proposal_state.proposal_count),
                                       proposal_state.proposal_count)

            enough_proposals = proposal_count >= self.num_slices

            # 0: successful proposal & enough proposals -> done
            # 1: successful proposal & not enough proposals -> change direction
            # 2: unsuccessful proposal -> shrink interval

            process_step = jnp.where(enough_proposals,
                                     jnp.full(good_proposal.shape, 0, int_type),
                                     jnp.where(good_proposal & ~enough_proposals,
                                               jnp.full(good_proposal.shape, 1, int_type),
                                               jnp.full(good_proposal.shape, 2, int_type)
                                               )
                                     )

            def _map_where(cond, a_tree, b_tree):
                return tree_map(lambda a, b: jnp.where(cond, a, b), a_tree, b_tree)

            proposal_state_from_1 = self._change_direction(proposal_state, log_L_point_U)
            proposal_state_from_2 = self._shrink_interval(proposal_state, log_L_point_U,
                                                          log_L_contour=proposal_state.log_L_constraint,
                                                          midpoint_shrink=self.midpoint_shrink,
                                                          destructive_shrink=False)

            # replace with the proposal state as appropriate

            proposal_state = _map_where(process_step == 1,
                                        proposal_state_from_1, proposal_state)
            proposal_state = _map_where(process_step == 2,
                                        proposal_state_from_2, proposal_state)

            proposal_state = proposal_state._replace(process_step=process_step,
                                                     proposal_count=proposal_count,
                                                     num_likelihood_evaluations=num_likelihood_evaluations)
            return (proposal_state, log_L_point_U)

        def slice_sampler_cond(body_state: CarryType) -> BoolArray:
            """
            Stops when there have been enough proposals.
            """
            (proposal_state, _) = body_state
            return jnp.less(proposal_state.proposal_count, self.num_slices)

        (proposal_state, log_L) = while_loop(slice_sampler_cond,
                                             slice_sampler_body,
                                             (init_proposal_state, -jnp.inf))
        # when num_slices==0, that means we don't want to run this
        pass_through = self.num_slices == jnp.zeros_like(self.num_slices)
        log_L = jnp.where(pass_through, seed_point.log_L0, log_L)
        point_U = jnp.where(pass_through, seed_point.U0, proposal_state.point_U)
        sample = Sample(point_U=point_U,
                        log_L_constraint=proposal_state.log_L_constraint,
                        log_L=log_L,
                        num_likelihood_evaluations=proposal_state.num_likelihood_evaluations,
                        num_slices=self.num_slices,
                        iid=jnp.asarray(False, jnp.bool_))
        return sample


class MultiDimProposalState(NamedTuple):
    key: jnp.ndarray  # PRNG key
    process_step: jnp.ndarray  # what stage of the sampling each independent point is at
    proposal_count: jnp.ndarray  # how many successful proposals have happened for each independent point
    num_likelihood_evaluations: jnp.ndarray  # how many  likelihood evaluations have been made for each independent point
    point_U0: jnp.ndarray  # the origin of current slice of each independent point
    log_L0: jnp.ndarray  # the value of log-likelihood at each indpendent point origin
    left: jnp.ndarray  # the left bound of slice
    right: jnp.ndarray  # the right bound of slice
    point_U: jnp.ndarray  # the point up for likelihood computation


class MultiDimSliceSampler(MarkovSampler):
    def __init__(self, model: Model, num_slices: int, num_restrict_dims: Optional[int] = None,
                 efficiency_threshold: Optional[float] = None):
        """
        Multi-dimensional slice sampler, with exponential shrinkage. Produces correlated (non-i.i.d.) samples.

        Notes: Not very efficient.

        Args:
            model: Model
            num_slices: number of slices between acceptance, in units of 1, unlike other software which does it in units of prior dimension.
            num_restrict_dims: size of subspace to slice along. Setting to 1 would be like UniDimSliceSampler,
                but far less efficient.
        """
        super().__init__(model=model, efficiency_threshold=efficiency_threshold)
        if num_slices < 1:
            raise ValueError(f"num_slices must be > 0.")
        self.num_slices = num_slices
        if num_restrict_dims is not None:
            if num_restrict_dims == 1:
                raise ValueError(f"If restricting to 1 dimension, then you should use UniDimSliceSampler.")
            if not (1 < num_restrict_dims <= model.U_ndims):
                raise ValueError(f"Expected num_restriction dim in (1, {model.U_ndims}], got {num_restrict_dims}.")
        self.num_restrict_dims = num_restrict_dims

    def preprocess(self, state: NestedSamplerState) -> PreProcessType:
        return ()

    def _slice_bounds(self, key: PRNGKey, point_U0: FloatArray) -> Tuple[FloatArray, FloatArray]:
        """
        Get the slice bounds, randomly selecting which dimensions to slice in.

        Args:
            key: PRNGKey
            point_U0: the seed point

        Returns:
            left, and right bounds of slice
        """
        if self.num_restrict_dims is not None:
            slice_dims = random.choice(key, point_U0.size, shape=(self.num_restrict_dims,), replace=False)
            left = point_U0.at[slice_dims].set(jnp.zeros(self.num_restrict_dims, point_U0.dtype))
            right = point_U0.at[slice_dims].set(jnp.ones(self.num_restrict_dims, point_U0.dtype))
        else:
            left = jnp.zeros_like(point_U0)
            right = jnp.ones_like(point_U0)
        return left, right

    def _new_slice(self, proposal_state: MultiDimProposalState, log_L_U: FloatArray) -> MultiDimProposalState:
        key, slice_key, sample_key = random.split(proposal_state.key, 3)

        left, right = self._slice_bounds(key=slice_key, point_U0=proposal_state.point_U0)

        point_U = random.uniform(key=sample_key, shape=left.shape, dtype=left.dtype, minval=left, maxval=right)

        next_proposal_state = MultiDimProposalState(
            key=key,
            process_step=jnp.full((), 3, int_type),
            proposal_count=proposal_state.proposal_count,
            num_likelihood_evaluations=proposal_state.num_likelihood_evaluations,
            point_U0=proposal_state.point_U,
            log_L0=log_L_U,
            left=left,
            right=right,
            point_U=point_U
        )
        return next_proposal_state

    def _shrink_region(self, proposal_state: MultiDimProposalState) -> MultiDimProposalState:
        # if point_U is on the 'right' side then we shrink the 'right' side to it.
        # same of 'left'
        left = jnp.where(proposal_state.point_U < proposal_state.point_U0,
                         jnp.maximum(proposal_state.left, proposal_state.point_U),
                         proposal_state.left)
        right = jnp.where(proposal_state.point_U > proposal_state.point_U0,
                          jnp.minimum(proposal_state.right, proposal_state.point_U),
                          proposal_state.right)

        key, sample_key = random.split(proposal_state.key, 2)

        point_U = random.uniform(key=sample_key, shape=left.shape, dtype=left.dtype, minval=left, maxval=right)

        next_proposal_state = proposal_state._replace(
            key=key,
            process_step=jnp.full((), 3, int_type),
            left=left,
            right=right,
            point_U=point_U
        )

        return next_proposal_state

    def get_sample_from_seed(self, key: PRNGKey, seed_point: SeedPoint, log_L_constraint: FloatArray,
                             preprocess_data: PreProcessType) -> Sample:
        slice_key, sample_key, proposal_key = random.split(key, 3)

        num_likelihood_evaluations = jnp.full((), 0, int_type)

        left, right = self._slice_bounds(key=slice_key, point_U0=seed_point.U0)

        point_U = random.uniform(key=sample_key, shape=left.shape, dtype=left.dtype, minval=left, maxval=right)

        init_proposal_state = MultiDimProposalState(key=proposal_key,
                                                    process_step=jnp.full((), 3, int_type),
                                                    proposal_count=jnp.zeros((), int_type),
                                                    num_likelihood_evaluations=num_likelihood_evaluations,
                                                    point_U0=seed_point.U0,
                                                    log_L0=seed_point.log_L0,
                                                    left=left,
                                                    right=right,
                                                    point_U=point_U)

        CarryType = Tuple[MultiDimProposalState, FloatArray]

        def slice_sampler_body(body_state: CarryType) -> CarryType:
            """
            Iteratively samples num_parallel_samplers new points using a sampler, then replaces that number of points in
            the reservoir.
            """
            (proposal_state, _) = body_state
            log_L_point_U = self.model.forward(proposal_state.point_U)
            num_likelihood_evaluations = proposal_state.num_likelihood_evaluations + jnp.ones_like(
                proposal_state.num_likelihood_evaluations)
            # assumes that log_L0 > log_L_constraint
            good_proposal = jnp.greater(log_L_point_U, log_L_constraint)
            proposal_count = jnp.where(good_proposal,
                                       proposal_state.proposal_count + jnp.ones_like(proposal_state.proposal_count),
                                       proposal_state.proposal_count)

            enough_proposals = proposal_count >= self.num_slices

            # 0: successful proposal & enough proposals -> done
            # 1: successful proposal & not enough proposals -> change direction
            # 2: unsuccessful proposal -> shrink interval

            process_step = jnp.where(good_proposal & enough_proposals,
                                     jnp.full(good_proposal.shape, 0, int_type),
                                     jnp.where(good_proposal & ~enough_proposals,
                                               jnp.full(good_proposal.shape, 1, int_type),
                                               jnp.full(good_proposal.shape, 2, int_type)
                                               )
                                     )

            def _map_where(cond, a_tree, b_tree):
                return tree_map(lambda a, b: jnp.where(cond, a, b), a_tree, b_tree)

            proposal_state_from_1 = self._new_slice(proposal_state, log_L_point_U)
            proposal_state_from_2 = self._shrink_region(proposal_state)

            # replace with the proposal state as appropriate

            proposal_state = _map_where(process_step == 1, proposal_state_from_1, proposal_state)
            proposal_state = _map_where(process_step == 2, proposal_state_from_2, proposal_state)

            proposal_state = proposal_state._replace(process_step=process_step,
                                                     proposal_count=proposal_count,
                                                     num_likelihood_evaluations=num_likelihood_evaluations)
            return (proposal_state, log_L_point_U)

        def slice_sampler_cond(body_state: CarryType) -> BoolArray:
            """
            Stops when there have been enough proposals.
            """
            (proposal_state, _) = body_state
            return jnp.bitwise_not(proposal_state.proposal_count >= self.num_slices)

        (proposal_state, log_L) = while_loop(slice_sampler_cond,
                                             slice_sampler_body,
                                             (init_proposal_state, -jnp.inf))
        sample = Sample(point_U=point_U,
                        log_L_constraint=log_L_constraint,
                        log_L=log_L,
                        num_likelihood_evaluations=proposal_state.num_likelihood_evaluations,
                        num_slices=self.num_slices,
                        iid=jnp.asarray(False, jnp.bool_))
        return sample



import logging
from typing import Tuple, NamedTuple, TypeVar, Optional, List

from etils.array_types import PRNGKey, IntArray, BoolArray, FloatArray
from jax import tree_map, numpy as jnp, random, pmap
from jax._src.lax.control_flow import while_loop, scan
from jax._src.lax.parallel import all_gather

from jaxns.model import Model
from jaxns.statistics import analyse_sample_collection
from jaxns.termination import determine_termination
from jaxns.types import NestedSamplerState, Reservoir, LivePoints, TerminationCondition, int_type, Sample, float_type
from jaxns.utils import collect_samples

logger = logging.getLogger('jaxns')

__all__ = ['AbstractSampler', 'MarkovSampler', 'AbstractNestedSampler', 'StaticNestedSampler']

T = TypeVar('T')


def remove_chunk_dim(py_tree: T) -> T:
    def _remove_chunk_dim(a):
        shape = list(a.shape)
        shape = [shape[0] * shape[1]] + shape[2:]
        return jnp.reshape(a, shape)

    return tree_map(_remove_chunk_dim, py_tree)


def add_chunk_dim(py_tree: T, chunk_size: int) -> T:
    def _add_chunk_dim(a):
        shape = list(a.shape)
        shape = [chunk_size, shape[0] // chunk_size] + shape[1:]
        return jnp.reshape(a, shape)

    return tree_map(_add_chunk_dim, py_tree)


PreProcessType = TypeVar('PreProcessType')


class SeedPoint(NamedTuple):
    U0: FloatArray
    log_L0: FloatArray


class AbstractSampler:
    def __init__(self, model: Model, efficiency_threshold: Optional[FloatArray] = None):
        self.model = model
        if efficiency_threshold is None:
            efficiency_threshold = 0.
        if efficiency_threshold < 0. or efficiency_threshold >= 1.:
            raise ValueError(f"{efficiency_threshold} must be in [0., 1.), got {efficiency_threshold}.")
        efficiency_threshold = jnp.asarray(efficiency_threshold, float_type)
        self.efficiency_threshold = efficiency_threshold

    def __repr__(self):
        return self.__class__.__name__

    def preprocess(self, state: NestedSamplerState) -> PreProcessType:
        """
        Produces a data structure that is necessary for sampling to run.
        Typically this is where clustering happens.

        Args:
            state: nested sampler state

        Returns:
            any valid pytree
        """
        raise NotImplementedError()

    def get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, live_points: LivePoints,
                   preprocess_data: PreProcessType) -> Sample:
        """
        Produce a single i.i.d. sample from the model within the log_L_constraint.

        Args:
            key: PRNGkey
            log_L_constraint: the constraint to sample within
            live_points: the current live points reservoir
            preprocess_data: the data pytree needed and produced by the sampler

        Returns:
            an i.i.d. sample
        """
        raise NotImplementedError()


class RejectionSampler(AbstractSampler):
    """
    Samplers that are based on rejection sampling. They usually first-lines of attack, and are stopped once efficiency
    gets too low.
    """
    pass


class MarkovSampler(AbstractSampler):
    """
    A sampler that conditions off a known satisfying point, e.g. a seed point.
    """

    def get_seed_point(self, key: PRNGKey, live_points: LivePoints, log_L_constraint: FloatArray) -> SeedPoint:
        """
        Samples a seed point from the live points.

        Args:
            key: PRNGKey
            live_points: the current live point set. All points satisfy the log-L constraint
            log_L_constraint: a log-L constraint to sample within. Note: Currently, redundant because we assume live
                points satisfies the constraint, but in the future, some points may not and this will be used.

        Returns:
            a seed point
        """
        sample_idx = random.randint(key, (), minval=0, maxval=live_points.reservoir.log_L.size)
        return SeedPoint(
            U0=live_points.reservoir.point_U[sample_idx],
            log_L0=live_points.reservoir.log_L[sample_idx]
        )

    def get_sample_from_seed(self, key: PRNGKey, seed_point: SeedPoint, log_L_constraint: FloatArray,
                             preprocess_data: PreProcessType) -> Sample:
        """
        Produce a single i.i.d. sample from the model within the log_L_constraint.

        Args:
            key: PRNGkey
            seed_point: function that gets the next sample from a seed point
            log_L_constraint: the constraint to sample within
            preprocess_data: the data pytree needed and produced by the sampler

        Returns:
            an i.i.d. sample
        """
        raise NotImplementedError()

    def get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, live_points: LivePoints,
                   preprocess_data: PreProcessType) -> Sample:
        """
        Produce a single i.i.d. sample from the model within the log_L_constraint.

        Args:
            key: PRNGkey
            log_L_constraint: the constraint to sample within
            live_points: the current live points reservoir
            preprocess_data: the data pytree needed and produced by the sampler

        Returns:
            an i.i.d. sample
        """
        key, seed_key = random.split(key, 2)
        seed_point = self.get_seed_point(key=seed_key, live_points=live_points, log_L_constraint=log_L_constraint)
        return self.get_sample_from_seed(key=key, seed_point=seed_point, log_L_constraint=log_L_constraint,
                                         preprocess_data=preprocess_data)


class AbstractNestedSampler:
    def __call__(self, state: NestedSamplerState, live_points: LivePoints, termination_cond: TerminationCondition
                 ) -> Tuple[IntArray, NestedSamplerState, LivePoints]:
        """
        Performs nested sampling from an initial state with a static number of live points at each shrinkage step, and
        terminating upon ANY of the terminiation criteria being met.

        Args:
            state: the current state of sampler
            live_points: the live points to start sampling from
            termination_cond: the conditions for termination

        Returns:
            termination_reason, final state, and live points
        """
        raise NotImplementedError()


class StaticNestedSampler(AbstractNestedSampler):
    """
    Performs parallel static sampling up an underlying model
    """

    def __init__(self, samplers: List[AbstractSampler], num_live_points: int, num_parallel_samplers: int = 1):
        if num_live_points % num_parallel_samplers != 0:
            raise ValueError(
                f"num_live_points {num_live_points} must divide num_parallel_samplers {num_parallel_samplers}"
            )
        self.num_live_points = num_live_points
        self.num_parallel_samplers = num_parallel_samplers
        self.samplers = samplers

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.num_live_points} live points over {self.num_parallel_samplers} devices) using {'->'.join(map(repr, self.samplers))}"

    def _single_live_point_shrink(self,
                                  key: PRNGKey,
                                  live_points: LivePoints,
                                  log_L_contour: FloatArray,
                                  preprocess_data: PreProcessType,
                                  sampler: AbstractSampler) -> Tuple[Reservoir, LivePoints, FloatArray]:
        """
        Run nested sampling to replace an entire live point reservoir via shrinkage.

        Args:
            key: PRNGKey
            live_points: live points
            log_L_contour: the initial log-L contour of live-points.
                Note: this could be computed from the live_points reservoir but would be inefficient to do so.
            preprocess_data: any data needed to shrink the replace the live point reservoir with i.i.d. samples.
            sampler: sampler to use

        Returns:
            dead point reservoir, live points, the final log-L contour of live points
        """

        class CarryType(NamedTuple):
            key: PRNGKey
            live_points: LivePoints
            log_L_contour: FloatArray

        ResultType = Reservoir

        def body(carry: CarryType, unused_X: IntArray) -> Tuple[CarryType, ResultType]:
            del unused_X
            idx_min = jnp.argmin(carry.live_points.reservoir.log_L)
            dead_point: Reservoir = tree_map(lambda x: x[idx_min], carry.live_points.reservoir)
            log_L_dead = dead_point.log_L

            # contour becomes log_L_dead if log_L_dead is not supremum of live-points, else we choose the original
            # constraint of dead point. Note: we are at liberty to choose any log_L level as a contour so long as we
            # can sample within it uniformly.

            next_dead_contour = jnp.min(carry.live_points.reservoir.log_L.at[idx_min].set(jnp.inf))
            on_supremum = jnp.equal(log_L_dead, next_dead_contour)
            log_L_contour = jnp.where(on_supremum, carry.log_L_contour, log_L_dead)

            # replace dead point with a new sample about contour
            key, sample_key = random.split(carry.key, 2)

            sample = sampler.get_sample(key=sample_key,
                                        log_L_constraint=log_L_contour,
                                        live_points=carry.live_points,
                                        preprocess_data=preprocess_data)

            live_points_reservoir = tree_map(lambda old, update: old.at[idx_min].set(update),
                                             carry.live_points.reservoir, Reservoir(*sample))
            live_points = carry.live_points._replace(reservoir=live_points_reservoir)

            return CarryType(key=key, live_points=live_points, log_L_contour=log_L_contour), dead_point

        init_carry = CarryType(key=key, live_points=live_points, log_L_contour=log_L_contour)
        init_X = live_points.reservoir.log_L
        (_, live_points, log_L_contour), dead_reservoir = scan(body, init_carry, init_X)
        return (dead_reservoir, live_points, log_L_contour)

    def _single_thread_ns(self, key: PRNGKey, state: NestedSamplerState, live_points: LivePoints,
                          termination_cond: TerminationCondition) -> Tuple[IntArray, NestedSamplerState, LivePoints]:
        """
        Runs a single thread of static nested sampling, using all-gather to compute stopping after each live-point
        set shrinkage, which is approximately equivalent to one e-fold decrease in enclosed prior volume. This
        continues until a stopping condition is reached. Due to the all-gather this stopping condition is based on
        all the data across all devices, and is the same on all devices.

        Args:
            key: PRNGKey
            state: the initial state
            live_points: the initial live points
            termination_cond: the termination condition

        Returns:
            termination_reason, state, live_points
        """
        state = state._replace(key=key)

        class CarryType(NamedTuple):
            done: BoolArray
            termination_reason: IntArray
            state: NestedSamplerState
            live_points: LivePoints
            log_L_contour: FloatArray

        def stopping_cond(state: NestedSamplerState, live_points: LivePoints, term_cond: TerminationCondition) -> Tuple[
            BoolArray, IntArray]:
            # Collect all live points from all devices to compute stopping condition
            all_live_points: LivePoints = remove_chunk_dim(all_gather(live_points, 'i'))
            evidence_calculation, sample_stats = analyse_sample_collection(
                sample_collection=state.sample_collection,
                sorted_collection=True
            )
            done, termination_reason = determine_termination(
                term_cond=term_cond,
                sample_collection=state.sample_collection,
                evidence_calculation=evidence_calculation,
                live_points=all_live_points
            )
            return done, termination_reason

        def build_body(sampler: AbstractSampler, term_cond: TerminationCondition):
            def body(body_state: CarryType) -> CarryType:
                key, sample_key = random.split(body_state.state.key, 2)
                state = body_state.state._replace(key=key)

                preprocess_data = sampler.preprocess(state)

                (dead_reservoir, live_points, log_L_contour) = self._single_live_point_shrink(
                    key=sample_key,
                    live_points=body_state.live_points,
                    log_L_contour=body_state.log_L_contour,
                    preprocess_data=preprocess_data,
                    sampler=sampler
                )
                # Collect dead reservoirs from all devices, and merge into state
                all_dead_reservoir: Reservoir = remove_chunk_dim(all_gather(dead_reservoir, 'i'))
                all_live_points: LivePoints = remove_chunk_dim(all_gather(live_points, 'i'))
                new_state = collect_samples(state, all_dead_reservoir)

                done, termination_reason = stopping_cond(state=new_state, live_points=all_live_points, term_cond=term_cond)

                return CarryType(done=done, termination_reason=termination_reason, state=new_state,
                                 live_points=live_points, log_L_contour=log_L_contour)

            return body

        # We start with live points sorted. The starting contour is the constraint of the minimal sample.
        sort_idx = jnp.lexsort((live_points.reservoir.log_L_constraint, live_points.reservoir.log_L))
        live_points: LivePoints = tree_map(lambda x: x[sort_idx], live_points)
        log_L_contour = live_points.reservoir.log_L_constraint[0]

        carry_state = CarryType(done=jnp.asarray(False, jnp.bool_),
                                termination_reason=jnp.asarray(0, int_type),
                                state=state, live_points=live_points, log_L_contour=log_L_contour)
        for sampler in self.samplers:
            # Set efficiency termination point
            term_cond = termination_cond._replace(efficiency_threshold=sampler.efficiency_threshold)
            done, termination_reason = stopping_cond(
                state=carry_state.state,
                live_points=carry_state.live_points,
                term_cond=term_cond
            )

            carry_state = carry_state._replace(
                done=done
            )

            carry_state = while_loop(
                lambda body_state: jnp.bitwise_not(body_state.done),
                build_body(sampler=sampler, term_cond=term_cond),
                carry_state
            )

        return carry_state.termination_reason, carry_state.state, carry_state.live_points

    def __call__(self, state: NestedSamplerState, live_points: LivePoints, termination_cond: TerminationCondition
                 ) -> Tuple[IntArray, NestedSamplerState, LivePoints]:
        if live_points.reservoir.log_L.size != self.num_live_points:
            raise ValueError(
                f"live points reservoir is the wrong size. "
                f"Got {live_points.reservoir.log_L.size} but expected {self.num_live_points}.")

        # live_points should be randomly shuffled between devices
        chunked_live_points = add_chunk_dim(live_points, self.num_parallel_samplers)

        keys = random.split(state.key, self.num_parallel_samplers)

        parallel_ns = pmap(lambda key, live_points: self._single_thread_ns(
            key=key,
            state=state,
            live_points=live_points,
            termination_cond=termination_cond
        ), axis_name='i')

        chunked_termination_reason, chunked_state, chunked_live_points = parallel_ns(keys, chunked_live_points)

        termination_reason, state = tree_map(lambda x: x[0], (chunked_termination_reason, chunked_state))
        live_points = remove_chunk_dim(chunked_live_points)

        return termination_reason, state, live_points



from typing import NamedTuple, Optional, Union, Any, Callable, Tuple, Dict

from etils.array_types import FloatArray, IntArray, PRNGKey, BoolArray
from jax import numpy as jnp

__all__ = ['Sample',
           'Reservoir',
           'SampleStatistics',
           'SampleCollection',
           'EvidenceCalculation',
           'NestedSamplerState',
           'LivePoints',
           'TerminationCondition',
           'NestedSamplerResults']

float_type = jnp.result_type(float)
int_type = jnp.result_type(int)
complex_type = jnp.result_type(complex)

LikelihoodType = Callable[..., FloatArray]
LikelihoodInputType = Tuple[jnp.ndarray, ...]  # Likeihood conditional variables
UType = FloatArray  # Homogeneous measure samples
XType = Dict[str, jnp.ndarray]  # Prior sample


class Sample(NamedTuple):
    """
    Holds the reservoir of new samples before merging.
    """
    # TODO: allow any pytree for points_X
    point_U: UType  # [d] sample in U-space
    log_L_constraint: FloatArray  # constraint that sample was sampled uniformly within
    log_L: FloatArray  # log likelihood of the sample
    num_likelihood_evaluations: IntArray  # how many times the likelihood was evaluated to produce this sample
    num_slices: IntArray  # the number of slices for sliced points.
    iid: BoolArray  # whether the sample is exactly iid sampled from within the likelihood constraint


class Reservoir(NamedTuple):
    """
    Holds the reservoir samples, has a leading [n] dimension.
    """
    # TODO: allow any pytree for points_X
    point_U: UType  # [d] sample in U-space
    log_L_constraint: FloatArray  # constraint that sample was sampled uniformly within
    log_L: FloatArray  # log likelihood of the sample
    num_likelihood_evaluations: IntArray  # how many times the likelihood was evaluated to produce this sample
    num_slices: IntArray  # the number of slices for sliced points.
    iid: BoolArray  # whether the sample is exactly iid sampled from within the likelihood constraint


class SampleStatistics(NamedTuple):
    num_live_points: IntArray  # [max_samples] How many live points were taken for the samples.
    log_dZ_mean: FloatArray  # [max_samples] The log mean evidence difference of the sample
    log_X_mean: FloatArray  # [max_samples] The log mean enclosed prior volume of sample


class SampleCollection(NamedTuple):
    """
    Arrays to hold samples taken from the reservoir.
    """

    sample_idx: IntArray  # the sample index, pointing to the next empty sample slot
    reservoir: Reservoir  # reservoir of samples with leading dimension [max_samples]


class EvidenceCalculation(NamedTuple):
    """
    Contains a running estimate of evidence and related quantities.
    """
    log_X_mean: FloatArray
    log_X2_mean: FloatArray
    log_Z_mean: FloatArray
    log_ZX_mean: FloatArray
    log_Z2_mean: FloatArray
    log_dZ2_mean: FloatArray


class NestedSamplerState(NamedTuple):
    key: PRNGKey
    sample_collection: SampleCollection  # Holds all collected samples so far.


class LivePoints(NamedTuple):
    reservoir: Reservoir


class TerminationCondition(NamedTuple):
    ess: Optional[FloatArray] = jnp.inf
    evidence_uncert: Optional[FloatArray] = jnp.asarray(0., float_type)
    live_evidence_frac: Optional[FloatArray] = jnp.asarray(1e-4, float_type)
    max_samples: Optional[IntArray] = jnp.iinfo(int_type).max
    max_num_likelihood_evaluations: Optional[IntArray] = jnp.iinfo(int_type).max
    log_L_contour: Optional[FloatArray] = jnp.inf
    efficiency_threshold: Optional[FloatArray] = jnp.asarray(0., float_type)


class NestedSamplerResults(NamedTuple):
    log_Z_mean: FloatArray  # estimate of E[log(Z)]
    log_Z_uncert: FloatArray  # estimate of StdDev[log(Z)]
    ESS: FloatArray  # estimate of Kish's effective sample size
    H_mean: FloatArray  # estimate of E[int log(L) L dp/Z]
    samples: XType  # Dict of arrays with leading dimension num_samples
    log_L_samples: FloatArray  # log(L) of each sample
    log_dp_mean: FloatArray  # log(E[dZ]) of each sample, where dZ is how much it contributes to the total evidence.
    log_X_mean: FloatArray  # log(E[U]) of each sample
    log_posterior_density: FloatArray  # log(P( theta | D )) log posteriori density
    num_live_points_per_sample: IntArray  # how many live points were taken for the samples.
    num_likelihood_evaluations_per_sample: IntArray  # how many likelihood evaluations were made per sample.
    num_slices: IntArray  # how many slices were taken for slice sampled points
    total_num_samples: IntArray  # int, the total number of samples collected.
    total_num_slices: IntArray  # int, how many slices in total were taken
    total_num_likelihood_evaluations: IntArray  # how many likelihood evaluations were made in total
    log_efficiency: FloatArray  # log(total_num_samples / total_num_likelihood_evaluations)
    termination_reason: IntArray  # this will be an int reflecting the reason for termination


class SignedLog(NamedTuple):
    """
    Represents a signed value in log-space
    """
    log_abs_val: jnp.ndarray
    sign: Union[jnp.ndarray, Any]



# import tensorflow_probability.substrates.jax as tfp
# from jax import random, numpy as jnp
#
# from jaxns import collect_samples
# from jaxns.internals.log_semiring import LogSpace
# from jaxns.model import Model
# from jaxns.nested_sampler import ApproximateNestedSampler
# from jaxns.prior import PriorModelGen, Prior
# from jaxns.static_uniform import BadUniformSampler
# from jaxns.tests.test_nested_sampler import compute_shrinkage_stats
# from jaxns.types import TerminationCondition
#
# tfpd = tfp.distributions
#
#
# def test_adaptive_refinement():
#     n = 2
#
#     # Prior is uniform in U[0,1]
#     # Likelihood is 1 - x**n
#     # Z = 1 - 1/n+1
#
#     log_Z_true = jnp.log(1. - 1. / (n + 1))
#     print(f"True log(Z): {log_Z_true}")
#
#     def prior_model() -> PriorModelGen:
#         x = yield Prior(tfpd.Uniform(low=0, high=1))
#         return x
#
#     def log_likelihood(x):
#         return (LogSpace(0.) - LogSpace(n * jnp.log(x))).log_abs_val
#
#     def exact_X(L):
#         return (1. - L) ** (1. / n)
#
#     def exact_L(X):
#         return 1. - X ** n
#
#     model = Model(prior_model=prior_model,
#                   log_likelihood=log_likelihood)
#     ns = ApproximateNestedSampler(
#         model=model,
#         num_live_points=50,
#         num_parallel_samplers=1,
#         max_samples=1e6,
#         sampler_chain=[
#             BadUniformSampler(mis_fraction=0., model=model)
#         ]
#     )
#
#     total_state = None
#     for seed in [42, 43, 44]:
#         termination_reason, state = ns(random.PRNGKey(seed),
#                                        term_cond=TerminationCondition(live_evidence_frac=1e-4))
#         termination_reason.block_until_ready()
#         if total_state is None:
#             total_state = state
#         else:
#             total_state = collect_samples(state=total_state, new_reservoir=state.sample_collection.reservoir)
#     results = ns.to_results(total_state, termination_reason)
#     ns.summary(results)
#     ns.plot_diagnostics(results)
#
#     # ensure there is no bug in control flow. X should be same to controlled evaluation
#     log_X_mean, log_X_std = compute_shrinkage_stats(results.num_live_points_per_sample)
#     assert jnp.allclose(results.log_X_mean, log_X_mean)
#
#     # ensure the deviation from the exact shrinkage is correct
#     X_exact = exact_X(jnp.exp(results.log_L_samples))
#
#     rel_diff = jnp.abs(jnp.exp(log_X_mean) - X_exact) / jnp.exp(log_X_std)
#
#     # ensure log_Z is close to truth
#     assert jnp.isclose(results.log_Z_mean, log_Z_true, atol=results.log_Z_uncert * 1.75)
#
#     print("Relative shrinkage errors", jnp.percentile(rel_diff, jnp.asarray([50, 75, 90, 95])))
#     assert jnp.all(jnp.percentile(rel_diff, jnp.asarray([50, 75, 90, 95])) < jnp.asarray([0.9, 1.1, 1.4, 1.5]))



import logging
from functools import partial
from typing import Optional, Tuple, Union, List

import tensorflow_probability.substrates.jax as tfp
from etils.array_types import PRNGKey, IntArray
from jax import random, numpy as jnp, core, tree_map, vmap, jit

from jaxns.adaptive_refinement import AdaptiveRefinement
from jaxns.initial_state import init_sample_collection, get_uniform_init_live_points
from jaxns.internals.log_semiring import LogSpace, normalise_log_space
from jaxns.internals.stats import linear_to_log_stats, effective_sample_size
from jaxns.model import Model
from jaxns.plotting import plot_cornerplot, plot_diagnostics
from jaxns.static_nested_sampler import StaticNestedSampler, AbstractSampler
from jaxns.static_slice import UniDimSliceSampler
from jaxns.static_uniform import UniformSampler
from jaxns.statistics import analyse_sample_collection
from jaxns.types import TerminationCondition, NestedSamplerState, NestedSamplerResults, LivePoints
from jaxns.utils import collect_samples
from jaxns.utils import summary, save_results, load_results

tfpd = tfp.distributions

logger = logging.getLogger('jaxns')

__all__ = ['BaseNestedSampler',
           'ApproximateNestedSampler',
           'ExactNestedSampler']


class BaseNestedSampler:
    def __init__(self, model: Model, max_samples: Union[int, float]):
        self.model = model
        self.max_samples = int(max_samples)

    def initialise(self, key: PRNGKey, num_live_points: int) -> Tuple[NestedSamplerState, LivePoints]:
        """
        Creates initial live points from -inf contour and state.

        Args:
            key: PRNGKey

        Returns:
            initial state, and live points
        """
        state_key, live_points_key = random.split(key, 2)
        init_state = NestedSamplerState(
            key=state_key,
            sample_collection=init_sample_collection(size=self.max_samples, model=self.model)
        )
        live_points = get_uniform_init_live_points(live_points_key,
                                                   num_live_points=num_live_points,
                                                   model=self.model)
        return init_state, live_points

    def resize_state(self, state: NestedSamplerState, max_num_samples: int) -> NestedSamplerState:
        """
        Grow the state to the size of max_num_samples.

        Args:
            state: state to enlarge
            max_num_samples: size of resulting sample collection

        Returns:
            resized state
        """
        reservoir = state.sample_collection.reservoir
        if max_num_samples <= reservoir.log_L.size:
            logger.warning(
                f"expected max_num_samples larger than the current size {reservoir.log_L.size}")
            return state
        diff_size = max_num_samples - reservoir.log_L.size
        if diff_size <= 0:
            return state
        reservoir_extra = init_sample_collection(size=diff_size, model=self.model).reservoir
        reservoir = tree_map(lambda old, update: jnp.concatenate([old, update], axis=0),
                             reservoir,
                             reservoir_extra)
        return state._replace(sample_collection=state.sample_collection._replace(reservoir=reservoir))

    def summary(self, results: NestedSamplerResults) -> str:
        return summary(results)

    def plot_cornerplot(self, results: NestedSamplerResults, vars: Optional[List[str]] = None):
        plot_cornerplot(results, vars=vars)

    def plot_diagnostics(self, results: NestedSamplerResults):
        plot_diagnostics(results)

    def save_results(self, results: NestedSamplerResults, save_file: str):
        save_results(results, save_file)

    def load_results(self, save_file: str):
        return load_results(save_file)

    def to_results(self, state: NestedSamplerState, termination_reason: IntArray) -> NestedSamplerResults:
        """
        Converts a state into a result.

        Args:
            state: state
            termination_reason: the termination reason

        Returns:
            results object
        """

        if isinstance(state.sample_collection.reservoir.iid, core.Tracer):
            raise RuntimeError("Tracer detected, but expected imperative context.")

        num_samples = jnp.minimum(state.sample_collection.sample_idx, state.sample_collection.reservoir.log_L.size)

        sample_collection = state.sample_collection._replace(
            reservoir=tree_map(lambda x: x[:num_samples], state.sample_collection.reservoir)
        )

        evidence_calculation, sample_stats = analyse_sample_collection(
            sample_collection=sample_collection,
            sorted_collection=True
        )

        log_Z_mean, log_Z_var = linear_to_log_stats(
            log_f_mean=evidence_calculation.log_Z_mean,
            log_f2_mean=evidence_calculation.log_Z2_mean
        )
        log_Z_uncert = jnp.sqrt(log_Z_var)

        # Kish's ESS = [sum dZ]^2 / [sum dZ^2]
        ESS = effective_sample_size(evidence_calculation.log_Z_mean, evidence_calculation.log_dZ2_mean)

        samples = vmap(self.model.transform)(sample_collection.reservoir.point_U)

        log_L_samples = sample_collection.reservoir.log_L
        dp_mean = LogSpace(sample_stats.log_dZ_mean)
        dp_mean = normalise_log_space(dp_mean)
        H_mean = LogSpace(jnp.where(jnp.isneginf(dp_mean.log_abs_val),
                                    -jnp.inf,
                                    dp_mean.log_abs_val + log_L_samples)).sum().value - log_Z_mean
        X_mean = LogSpace(sample_stats.log_X_mean)
        num_likelihood_evaluations_per_sample = sample_collection.reservoir.num_likelihood_evaluations
        total_num_likelihood_evaluations = jnp.sum(num_likelihood_evaluations_per_sample)
        num_live_points_per_sample = sample_stats.num_live_points
        efficiency = LogSpace(jnp.log(num_samples) - jnp.log(total_num_likelihood_evaluations))

        log_posterior_density = sample_collection.reservoir.log_L + vmap(self.model.log_prob_prior)(
            sample_collection.reservoir.point_U)

        total_num_slices = jnp.sum(sample_collection.reservoir.num_slices)

        return NestedSamplerResults(
            log_Z_mean=log_Z_mean,  # estimate of log(E[Z])
            log_Z_uncert=log_Z_uncert,  # estimate of log(StdDev[Z])
            ESS=ESS,  # estimate of Kish's effective sample size
            H_mean=H_mean,  # estimate of E[int log(L) L dp/Z]
            total_num_samples=num_samples,  # int, the total number of samples collected.
            log_L_samples=log_L_samples,  # log(L) of each sample
            log_dp_mean=dp_mean.log_abs_val,
            log_posterior_density=log_posterior_density,
            # log(E[dZ]) of each sample, where dZ is how much it contributes to the total evidence.
            # log(StdDev[dZ]) of each sample, where dZ is how much it contributes to the total evidence.
            log_X_mean=X_mean.log_abs_val,  # log(E[U]) of each sample
            num_likelihood_evaluations_per_sample=num_likelihood_evaluations_per_sample,
            # how many likelihood evaluations were made per sample.
            num_live_points_per_sample=num_live_points_per_sample,
            # how many live points were taken for the samples.
            total_num_likelihood_evaluations=total_num_likelihood_evaluations,
            # how many likelihood evaluations were made in total,
            # sum of num_likelihood_evaluations_per_sample.
            total_num_slices=total_num_slices,
            log_efficiency=efficiency.log_abs_val,
            # total_num_samples / total_num_likelihood_evaluations
            termination_reason=termination_reason,  # termination condition as bit mask
            num_slices=sample_collection.reservoir.num_slices,
            samples=samples)

    def __call__(self, key: PRNGKey, term_cond: TerminationCondition, *,
                 init_state: Optional[NestedSamplerState] = None) -> Tuple[IntArray, NestedSamplerState]:
        """
        Performs approximate nested sampling followed by adaptive refinement.

        Args:
            key: PRNGKey
            term_cond: termination condition
            init_state: optional initial state

        Returns:
            termination reason, and exact state
        """
        raise NotImplementedError()


class ApproximateNestedSampler(BaseNestedSampler):
    """
    Performs nested sampling, using a chain of nested samplers, producing only weakly i.i.d. samples during the
    shrinkage process.
    """

    def __init__(self, model: Model, num_live_points: Union[int, float], num_parallel_samplers: int,
                 max_samples: Union[int, float],
                 sampler_chain: Optional[List[AbstractSampler]] = None):
        super().__init__(model=model, max_samples=max_samples)

        if sampler_chain is None:
            sampler_chain = [
                UniformSampler(model=model, efficiency_threshold=0.1),
                UniDimSliceSampler(model=model, num_slices=model.U_ndims * 3, midpoint_shrink=True, perfect=True)
            ]

        if sampler_chain[-1].efficiency_threshold is not None:
            if sampler_chain[-1].efficiency_threshold > 0.:
                logger.warning(
                    f"Your sampler chain will stop prematurely at an efficiency of "
                    f"{sampler_chain[-1].efficiency_threshold}. "
                    f"To change this behaviour, you should ensure the last sampler in your sampler_chain has no "
                    f"efficiency_threshold set."
                )

        remainder = int(num_live_points) % num_parallel_samplers
        extra = (num_parallel_samplers - remainder) % num_parallel_samplers
        if extra > 0:
            logger.warning(
                f"Increasing num_live_points ({num_live_points}) by {extra} to closest multiple of num_parallel_samplers.")
        num_live_points = int(num_live_points + extra)
        num_parallel_samplers = int(num_parallel_samplers)

        self._nested_sampler = StaticNestedSampler(samplers=sampler_chain,
                                                   num_live_points=num_live_points,
                                                   num_parallel_samplers=num_parallel_samplers)

    @property
    def num_live_points(self) -> int:
        return self._nested_sampler.num_live_points

    @property
    def nested_sampler(self) -> StaticNestedSampler:
        return self._nested_sampler

    def _run_chain(self, state: NestedSamplerState, live_points: LivePoints, term_cond: TerminationCondition) -> Tuple[
        IntArray, NestedSamplerState, LivePoints]:

        termination_reason, state, live_points = self._nested_sampler(
            state=state, live_points=live_points, termination_cond=term_cond)
        return termination_reason, state, live_points

    def approximate_shrinkage(self, state: NestedSamplerState, live_points: LivePoints,
                              term_cond: TerminationCondition) -> Tuple[IntArray, NestedSamplerState]:
        """
        Performs approximate shrinkage using a fast but inaccurate static nested sampler.

        Note: samples are not sure to be i.i.d. from within the likelihood constraint.

        Args:
            state: initial state
            live_points: initial live points
            term_cond: termination condition

        Returns:
            termination reason, and state
        """
        termination_reason, state, live_points = self._nested_sampler(
            state=state, live_points=live_points, termination_cond=term_cond)
        state = collect_samples(state=state, new_reservoir=live_points.reservoir)
        return termination_reason, state

    @partial(jit, static_argnums=0)
    def fresh_run(self, key: PRNGKey, term_cond: TerminationCondition) -> Tuple[IntArray, NestedSamplerState]:
        """
        Creates a new initial state and live points from -inf contour and samples until termination.

        Args:
            key: PRNGKey
            term_cond: termination condition

        Returns:
            termination reason, and state
        """
        state, live_points = self.initialise(key=key, num_live_points=self.num_live_points)
        return self.approximate_shrinkage(state=state, live_points=live_points, term_cond=term_cond)

    @partial(jit, static_argnums=0)
    def concat_run(self, state: NestedSamplerState, live_points: LivePoints,
                   term_cond: TerminationCondition) -> Tuple[IntArray, NestedSamplerState, LivePoints]:
        """
        Uses a given live points reservoir and performs approximate sampling until termination with a given initial
        state.

        Args:
            state: state to augment onto
            live_points: live points to use
            term_cond: termination condition

        Returns:
            termination reason, state, and live points
        """
        termination_reason, state, live_points = self._nested_sampler(
            state=state, live_points=live_points, termination_cond=term_cond)
        return termination_reason, state, live_points

    @partial(jit, static_argnums=0)
    def augment_run(self, state: NestedSamplerState, term_cond: TerminationCondition) -> Tuple[
        IntArray, NestedSamplerState]:
        """
        Creates a new live points reservoir from -inf contour and performs approximate sampling until termination with
        a given initial state.

        Args:
            state: state to augment onto
            term_cond: termination condition

        Returns:
            termination reason, and state
        """
        key, live_points_key = random.split(state.key)
        state = state._replace(key=key)
        live_points = get_uniform_init_live_points(live_points_key,
                                                   num_live_points=self.num_live_points,
                                                   model=self.model)
        return self.approximate_shrinkage(state=state, live_points=live_points, term_cond=term_cond)

    def __call__(self, key: PRNGKey, term_cond: TerminationCondition, *,
                 init_state: Optional[NestedSamplerState] = None) -> Tuple[IntArray, NestedSamplerState]:
        """
        Performs approximate nested sampling.

        Args:
            key: PRNGKey
            term_cond: termination conditions
            init_state: optional initial state, resume from here if provided

        Returns:
            termination reason, state
        """
        if term_cond.max_samples is None:  # add a convenience termination condition for user
            term_cond = term_cond._replace(max_samples=self.max_samples)
        else:  # truncate for sanity
            term_cond = term_cond._replace(max_samples=jnp.minimum(self.max_samples, term_cond.max_samples))
        if init_state is None:
            # We create fresh live points and state
            termination_reason, state = self.fresh_run(key=key, term_cond=term_cond)
        else:
            # We use the input live points and run on top of init_state
            init_state = self.resize_state(init_state, max_num_samples=self.max_samples)
            termination_reason, state = self.augment_run(state=init_state, term_cond=term_cond)
        return termination_reason, state


class ExactNestedSampler(BaseNestedSampler):
    """
    A two stage nested sampler, where the first stage produces sample that are only weakly i.i.d., and the second stage
    adaptively refines these samples until they are strongly i.i.d. (according to a stopping criterion).
    """

    def __init__(self, model: Model, num_live_points: Union[int, float],
                 max_samples: Union[int, float],
                 num_parallel_samplers: int = 1,
                 patience: int = 1):
        """
        A two stage nested sampler, where the first stage produces sample that are only weakly i.i.d., and the
        second stage adaptively refines these samples until they are strongly i.i.d. (according to a stopping
        criterion).

        Args:
            model: model
            num_live_points: the number of static live points
            max_samples: maximum number of (non equally weighted) samples to allocate space for.
                Note: to control stopping criterion related to a maximum number of samples, use TerminationCondition.
            num_parallel_samplers: the number of parallel instances to run. This must be <= len(jax.devices())
            patience: how many rounds of converged refinement to wait before stopping.
                Values > 1 mean refinement will continue until `patience` consecutive rounds of convergence are
                observed. Similar to famous patience parameter of early stopping in deep learning community.
        """
        super(ExactNestedSampler, self).__init__(model=model, max_samples=max_samples)

        self.approximate_sampler = ApproximateNestedSampler(
            model=model,
            num_live_points=num_live_points,
            num_parallel_samplers=num_parallel_samplers,
            max_samples=max_samples
        )

        self.adaptive_refinement = AdaptiveRefinement(
            model=self.model,
            patience=patience
        )

    def __call__(self, key: PRNGKey, term_cond: TerminationCondition, *,
                 init_state: Optional[NestedSamplerState] = None) -> Tuple[IntArray, NestedSamplerState]:
        """
        Performs approximate nested sampling followed by adaptive refinement.

        Args:
            key: PRNGKey
            term_cond: termination condition
            init_state: optional initial state

        Returns:
            termination reason, and exact state
        """
        if isinstance(key, core.Tracer):
            raise RuntimeError("Tracer detected, but expected imperative context.")

        termination_reason, state = self.approximate_sampler(key=key, term_cond=term_cond, init_state=init_state)
        state = self.adaptive_refinement(state=state)

        return termination_reason, state



from typing import Optional

from jax import numpy as jnp

from jaxns.internals.log_semiring import LogSpace
from jaxns.internals.stats import linear_to_log_stats, effective_sample_size
from jaxns.types import TerminationCondition, SampleCollection, EvidenceCalculation, LivePoints, \
    int_type

__all__ = ['determine_termination']


def determine_termination(term_cond: TerminationCondition,
                          sample_collection: Optional[SampleCollection] = None,
                          evidence_calculation: Optional[EvidenceCalculation] = None,
                          live_points: Optional[LivePoints] = None
                          ):
    """
    Determine if termination should happen.

    Args:
        term_cond: termination condition

    Returns:
        boolean done signal, and termination reason

    Termination Flags are bits:
        0-bit -> 1: used maximum allowed number of samples
        1-bit -> 2: evidence uncert below threshold
        2-bit -> 4: live points evidence below threshold
        3-bit -> 8: effective sample size big enough
        4-bit -> 16: used maxmimum allowed number of likelihood evaluations
        5-bit -> 32: maximum log-likelihood contour reached
        6-bit -> 64: sampler efficiency too low
        7-bit -> 128: entire live-points set is a single plateau

    Multiple flags are summed together
    """
    termination_condition = jnp.asarray(0, int_type)
    done = jnp.asarray(False)

    def _set_done_bit(bit, bit_reason, done, termination_condition):
        done = done | bit
        termination_condition += jnp.where(bit,
                                           jnp.asarray(2 ** bit_reason, int_type),
                                           jnp.asarray(0, int_type))
        return done, termination_condition

    if term_cond.max_samples is not None:
        if sample_collection is None:
            raise ValueError("sample_collections must not be None.")
        # used all points
        num_samples = sample_collection.sample_idx
        reached_max_samples = num_samples >= term_cond.max_samples
        done, termination_condition = _set_done_bit(reached_max_samples, 0,
                                                    done=done, termination_condition=termination_condition)
    if term_cond.evidence_uncert is not None:
        if evidence_calculation is None:
            raise ValueError("evidence_calculation must not be None.")
        _, log_Z_var = linear_to_log_stats(
            log_f_mean=evidence_calculation.log_Z_mean,
            log_f2_mean=evidence_calculation.log_Z2_mean)
        evidence_uncert_low_enough = log_Z_var <= jnp.square(term_cond.evidence_uncert)
        done, termination_condition = _set_done_bit(evidence_uncert_low_enough, 1,
                                                    done=done, termination_condition=termination_condition)
    if term_cond.live_evidence_frac is not None:
        if evidence_calculation is None:
            raise ValueError("evidence_calculation must not be None.")
        if live_points is None:
            raise ValueError("live_points must not be None.")
        # Z_remaining/(Z_remaining + Z_current) < delta => 1 + Z_current/Z_remaining > 1/delta
        remaining_evidence_upper_bound = LogSpace(jnp.max(live_points.reservoir.log_L)) * LogSpace(
            evidence_calculation.log_X_mean)
        log_Z_remaining_upper = remaining_evidence_upper_bound.log_abs_val

        evidence_upper_bound = remaining_evidence_upper_bound + LogSpace(evidence_calculation.log_Z_mean)
        log_Z_upper = evidence_upper_bound.log_abs_val

        small_remaining_evidence = log_Z_remaining_upper - log_Z_upper < jnp.log(term_cond.live_evidence_frac)
        done, termination_condition = _set_done_bit(small_remaining_evidence, 2,
                                                    done=done, termination_condition=termination_condition)
    if term_cond.ess is not None:
        if evidence_calculation is None:
            raise ValueError("evidence_calculation must not be None.")
        # Kish's ESS = [sum weights]^2 / [sum weights^2]
        ess = effective_sample_size(evidence_calculation.log_Z_mean,
                                    evidence_calculation.log_dZ2_mean)
        ess_reached = ess >= term_cond.ess
        done, termination_condition = _set_done_bit(ess_reached, 3,
                                                    done=done, termination_condition=termination_condition)
    if term_cond.max_num_likelihood_evaluations is not None:
        if sample_collection is None:
            raise ValueError("sample_collections must not be None.")
        num_likelihood_evaluations = jnp.sum(sample_collection.reservoir.num_likelihood_evaluations)
        too_max_likelihood_evaluations = num_likelihood_evaluations >= term_cond.max_num_likelihood_evaluations
        done, termination_condition = _set_done_bit(too_max_likelihood_evaluations, 4,
                                                    done=done, termination_condition=termination_condition)

    if term_cond.log_L_contour is not None:
        if sample_collection is None:
            raise ValueError("sample_collections must not be None.")
        log_L_max = jnp.max(
            jnp.where(jnp.arange(sample_collection.reservoir.log_L.size) < sample_collection.sample_idx,
                      sample_collection.reservoir.log_L,
                      -jnp.inf
                      )
        )
        likeihood_contour_reached = log_L_max >= term_cond.log_L_contour
        done, termination_condition = _set_done_bit(likeihood_contour_reached, 5,
                                                    done=done, termination_condition=termination_condition)

    if term_cond.efficiency_threshold is not None:
        if live_points is None:
            raise ValueError("live_points must be provided.")
        efficiency = jnp.reciprocal(jnp.mean(live_points.reservoir.num_likelihood_evaluations))
        efficiency_too_low = efficiency <= term_cond.efficiency_threshold
        done, termination_condition = _set_done_bit(efficiency_too_low, 6,
                                                    done=done, termination_condition=termination_condition)
    if (live_points is not None) and (live_points.reservoir.log_L.size > 1):
        all_plateau = jnp.min(live_points.reservoir.log_L) == jnp.max(live_points.reservoir.log_L)
        done, termination_condition = _set_done_bit(all_plateau, 7,
                                                    done=done, termination_condition=termination_condition)

    return done, termination_condition



import logging
from uuid import uuid4

import numpy as np
from etils.array_types import PRNGKey, FloatArray
from jax import random, vmap, jit, numpy as jnp

from jaxns.prior import PriorModelType, parse_prior, compute_log_likelihood, transform, log_prob_prior
from jaxns.types import float_type, LikelihoodType, UType, XType

logger = logging.getLogger('jaxns')

__all__ = ['Model']


class Model:
    """
    Represents a Bayesian model in terms of a generative prior, and likelihood function.
    """

    def __init__(self, prior_model: PriorModelType, log_likelihood: LikelihoodType):
        self.id = str(uuid4())
        self.prior_model = prior_model
        self.log_likelihood = log_likelihood
        self._U_placeholder, self._X_placeholder = parse_prior(prior_model=prior_model)
        self._U_ndims = self.U_placeholder.size

    @property
    def U_placeholder(self) -> UType:
        """
        A placeholder for U-space sample.
        """
        return self._U_placeholder

    @property
    def X_placeholder(self) -> XType:
        """
        A placeholder for X-space sample.
        """
        return self._X_placeholder

    @property
    def U_ndims(self) -> int:
        """
        The prior dimensionality.
        """
        return self._U_ndims

    def __hash__(self):
        return hash(self.id)

    def sample_U(self, key: PRNGKey) -> FloatArray:
        """
        Sample uniformly from the prior in U-space.

        Args:
            key: PRNGKey

        Returns:
            U-space sample
        """
        return random.uniform(key=key, shape=(self.U_ndims,), dtype=float_type)

    def transform(self, U: UType) -> XType:
        """
        Compute the prior sample.

        Args:
            U: U-space sample

        Returns:
            prior sample
        """
        return transform(U=U, prior_model=self.prior_model)

    def forward(self, U: UType, allow_nan: bool = False) -> FloatArray:
        """
        Compute the log-likelihood.

        Args:
            U: U-space sample
            allow_nan: whether to allow nans in likelihood

        Returns:
            log likelihood at the sample
        """
        return compute_log_likelihood(U=U, prior_model=self.prior_model, log_likelihood=self.log_likelihood,
                                      allow_nan=allow_nan)

    def log_prob_prior(self, U: UType) -> FloatArray:
        """
        Computes the log-probability of the prior.

        Args:
            U: The U-space sample

        Returns:
            the log probability of prior
        """
        return log_prob_prior(U=U, prior_model=self.prior_model)

    def sanity_check(self, key: PRNGKey, S: int):
        U = jit(vmap(self.sample_U))(random.split(key, S))
        log_L = jit(vmap(lambda u: self.forward(u, allow_nan=True)))(U)
        logger.info("Sanity check...")
        for _U, _log_L in zip(U, log_L):
            if jnp.isnan(_log_L):
                logger.info(f"Found bad point: {_U} -> {self.transform(_U)}")
        assert not any(np.isnan(log_L))
        logger.info("Sanity check passed")



from jax import random, numpy as jnp

from jaxns.utils import resample, _bit_mask


def test_resample():
    x = random.normal(key=random.PRNGKey(0), shape=(50,))
    logits = -jnp.ones(50)
    samples = {'x': x}
    assert jnp.all(resample(random.PRNGKey(0), samples, logits)['x'] == resample(random.PRNGKey(0), x, logits))


def test_bit_mask():
    assert _bit_mask(1, width=2) == [1, 0]
    assert _bit_mask(2, width=2) == [0, 1]
    assert _bit_mask(3, width=2) == [1, 1]



import numpy as np
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp, vmap, random

from jaxns import StaticNestedSampler, collect_samples
from jaxns.initial_state import init_sample_collection, get_uniform_init_live_points, \
    get_live_points_from_samples
from jaxns.model import Model
from jaxns.prior import PriorModelGen, Prior
from jaxns.static_uniform import UniformSampler
from jaxns.types import NestedSamplerState, SampleCollection, TerminationCondition

tfpd = tfp.distributions


def test_init_sample_collection():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z

    def log_likelihood(z):
        return -z ** 2

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)

    sample_collection = init_sample_collection(size=10, model=model)


def test_get_live_points_from_samples():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z

    def log_likelihood(z):
        return -z ** 2

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)

    U = jnp.stack(list(
        map(lambda x: x.flatten(), jnp.meshgrid(jnp.linspace(0., 1., 100), jnp.linspace(0., 1., 100), indexing='ij'))),
        axis=1)
    log_L = vmap(model.forward)(U)
    assert jnp.bitwise_not(jnp.any(jnp.isnan(log_L)))

    n = 100
    live_points = get_uniform_init_live_points(random.PRNGKey(43),
                                               num_live_points=n,
                                               model=model)
    init_state = NestedSamplerState(key=random.PRNGKey(42),
                                    sample_collection=SampleCollection(sample_idx=n,
                                                                       reservoir=live_points.reservoir))
    assert np.unique(init_state.sample_collection.reservoir.log_L).size == n
    state, live_points = get_live_points_from_samples(state=init_state,
                                                      log_L_constraint=-jnp.inf,
                                                      num_live_points=n - 1)

    unique = np.unique(live_points.reservoir.log_L)
    assert unique.size == n - 1

    assert np.unique(state.sample_collection.reservoir.log_L).size == 2  # 1 + empty type
    assert state.sample_collection.sample_idx == 1


def test_static_uniform():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z

    def log_likelihood(z):
        return -z ** 2

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)

    m = 1000
    sample_collection = init_sample_collection(size=m, model=model)

    state = NestedSamplerState(key=random.PRNGKey(42), sample_collection=sample_collection)

    n = 100
    live_points = get_uniform_init_live_points(random.PRNGKey(43),
                                               num_live_points=n,
                                               model=model)
    efficiency_threshold = 0.1
    ns = StaticNestedSampler(
        samplers=[UniformSampler(model=model, efficiency_threshold=efficiency_threshold)],
        num_live_points=n, num_parallel_samplers=1)
    _, state, live_points = ns(state=state, live_points=live_points,
                               termination_cond=TerminationCondition(max_samples=0))
    state = collect_samples(state=state, new_reservoir=live_points.reservoir)
    assert state.sample_collection.sample_idx > 0
    assert jnp.mean(1. / live_points.reservoir.num_likelihood_evaluations) > efficiency_threshold



"""Nested sampling with JAX."""
import logging

logger = logging.getLogger('jaxns')

logging.basicConfig(format='%(levelname)s[%(asctime)s]: %(message)s', level=logging.INFO)

from jaxns.nested_sampler import *
from jaxns.static_nested_sampler import *
from jaxns.model import *
from jaxns.types import *
from jaxns.plotting import *
from jaxns.prior import *
from jaxns.utils import *



import logging
from typing import Tuple, NamedTuple

from etils.array_types import PRNGKey, FloatArray, IntArray, BoolArray
from jax import core, numpy as jnp, tree_map, random, pmap
from jax._src.lax.control_flow import while_loop, scan
from jax._src.lax.parallel import all_gather

from jaxns.internals.stats import linear_to_log_stats
from jaxns.model import Model
from jaxns.static_nested_sampler import PreProcessType, add_chunk_dim, remove_chunk_dim, SeedPoint
from jaxns.static_slice import UniDimSliceSampler
from jaxns.statistics import analyse_sample_collection
from jaxns.types import NestedSamplerState, Reservoir, int_type, float_type
from jaxns.utils import sort_samples

__all__ = ['AdaptiveRefinement']

logger = logging.getLogger('jaxns')


class AdaptiveRefinement:
    """
    Class for adaptive refinement.
    """

    def __init__(self, model: Model, patience: int = 1, num_parallel_samplers: int = 1):
        """
        Initialised adaptive refinement.

        Args:
            model: Model
            patience: how many steps with stopping condition met before stopping
            num_parallel_samplers: how many parallel samplers to use.
        """
        if patience < 1:
            raise ValueError(f"patience should be >= 1, got {patience}.")
        self.patience = patience
        self.num_parallel_samplers = num_parallel_samplers
        self.sampler = UniDimSliceSampler(model=model, num_slices=model.U_ndims, midpoint_shrink=True, perfect=True)

    def _split_state(self, state: NestedSamplerState) -> Tuple[NestedSamplerState, Reservoir]:
        """
        Splice a state into iid samples and non-iid samples.

        Note: must not be run in a JIT-context, as the nonzero operation non-static.

        Args:
            state: state to split

        Returns:
            a state of iid samples, and reservoir of non-iid samples
        """
        if isinstance(state.sample_collection.reservoir.iid, core.Tracer):
            raise RuntimeError("Tracer detected, but expected imperative context.")
        num_samples = state.sample_collection.reservoir.log_L.size

        (update_indicies,) = jnp.nonzero(
            (jnp.bitwise_not(state.sample_collection.reservoir.iid) &
             (jnp.arange(num_samples) < state.sample_collection.sample_idx)
             )
        )

        (keep_indicies,) = jnp.nonzero(state.sample_collection.reservoir.iid)

        update_reservoir = tree_map(lambda x: x[update_indicies], state.sample_collection.reservoir)
        keep_reservoir = tree_map(lambda x: x[keep_indicies], state.sample_collection.reservoir)
        state = state._replace(
            sample_collection=state.sample_collection._replace(
                sample_idx=keep_indicies.size,
                reservoir=keep_reservoir
            )
        )
        return state, update_reservoir

    def _combine_state(self, state: NestedSamplerState, update_reservoir: Reservoir) -> NestedSamplerState:
        """
        Concatenates a state and reservoir and sorts the state.

        Args:
            state: state
            update_reservoir: reservoir

        Returns:
            merged and sorted state
        """
        # concatenate reservoirs
        reservoir = tree_map(lambda x, y: jnp.concatenate([x, y], axis=0),
                             state.sample_collection.reservoir, update_reservoir)

        sample_collection = state.sample_collection._replace(
            sample_idx=state.sample_collection.sample_idx + update_reservoir.iid.size,
            reservoir=reservoir
        )
        # Sort
        sample_collection = sort_samples(sample_collection)
        state = state._replace(sample_collection=sample_collection)
        return state

    def _single_improvement(self, key: PRNGKey, seed_point: SeedPoint, log_L_constraint: FloatArray,
                            preprocess_data: PreProcessType) -> Reservoir:
        """
        Perform perfect slice sampling, thereby improving the sample.

        Args:
            key: PRNGKey
            seed_point: seed point to sample from
            log_L_constraint: the constraint to sample inside
            preprocess_data: preprocessing data

        Returns:
            a sample from the seed point sampled within the given constraint
        """
        sample = self.sampler.get_sample_from_seed(key=key,
                                                   seed_point=seed_point,
                                                   log_L_constraint=log_L_constraint,
                                                   preprocess_data=preprocess_data)
        return Reservoir(*sample)

    def _single_improvement_batch(self,
                                  key: PRNGKey,
                                  non_iid_reservoir: Reservoir,
                                  preprocess_data: PreProcessType) -> Reservoir:
        """
        Run nested sampling to replace an entire live point reservoir via shrinkage.

        Args:
            key: PRNGKey
            non_iid_reservoir: reservoir to improve
            preprocess_data: any data needed to shrink the replace the live point reservoir with i.i.d. samples.

        Returns:
            improved reservoir
        """

        class CarryType(NamedTuple):
            key: PRNGKey

        class XsType(NamedTuple):
            reservoir_point: Reservoir

        ResultType = Reservoir

        def body(carry: CarryType, X: XsType) -> Tuple[CarryType, ResultType]:
            key, improve_key = random.split(carry.key, 2)
            seed_point = SeedPoint(U0=X.reservoir_point.point_U, log_L0=X.reservoir_point.log_L)
            improved_point = self._single_improvement(
                key=improve_key,
                seed_point=seed_point,
                log_L_constraint=X.reservoir_point.log_L_constraint,
                preprocess_data=preprocess_data
            )
            # Update stats
            improved_point = improved_point._replace(
                num_slices=improved_point.num_slices + X.reservoir_point.num_slices,
                num_likelihood_evaluations=(improved_point.num_likelihood_evaluations
                                            + X.reservoir_point.num_likelihood_evaluations)
            )

            return CarryType(key=key), improved_point

        _, improved_reservoir = scan(body,
                                     CarryType(key),
                                     XsType(reservoir_point=non_iid_reservoir))
        return improved_reservoir

    def _single_improve_thread(self, key: PRNGKey, iid_state: NestedSamplerState,
                               non_iid_reservoir: Reservoir) -> [NestedSamplerState, Reservoir]:
        """
        Performs perfect slice sampling on the set of samples that are not considered i.i.d. sampled within their
        respective likelihood constraint.

        Args:
            key: PRNGKey
            iid_state: state with only the i.i.d. samples
            non_iid_reservoir: samples that are not deemed i.i.d.

        Returns:
            reservoir where all samples are considered i.i.d. sampled within their respective likelihood constraints
        """

        iid_state = iid_state._replace(key=key)

        # update each sample one slice at a time until the stopping condition

        class CarryType(NamedTuple):
            key: PRNGKey
            non_iid_reservoir: Reservoir
            last_log_Z_mean: FloatArray
            last_diff_log_Z_mean: FloatArray
            converged_step_count: IntArray
            state: NestedSamplerState
            output_log_Z: FloatArray
            iter_j: IntArray

        def cond(body_state: CarryType) -> BoolArray:
            done = jnp.greater_equal(body_state.converged_step_count, self.patience)
            done = body_state.iter_j >= 0  # TODO: remove
            return jnp.bitwise_not(done)

        def body(body_state: CarryType) -> CarryType:
            key, improve_key = random.split(body_state.state.key, 2)
            state = body_state.state._replace(key=key)

            preprocess_data = self.sampler.preprocess(state)

            update_reservoir = self._single_improvement_batch(
                key=improve_key,
                non_iid_reservoir=body_state.non_iid_reservoir,
                preprocess_data=preprocess_data
            )

            all_update_reservoir: Reservoir = remove_chunk_dim(all_gather(update_reservoir, 'i'))

            all_state = self._combine_state(state=iid_state, update_reservoir=all_update_reservoir)

            evidence_calculation, sample_stats = analyse_sample_collection(
                sample_collection=all_state.sample_collection,
                sorted_collection=True
            )
            log_Z_mean, log_Z_var = linear_to_log_stats(
                log_f_mean=evidence_calculation.log_Z_mean,
                log_f2_mean=evidence_calculation.log_Z2_mean
            )
            nan_last_log_Z_mean = jnp.isnan(body_state.last_log_Z_mean)
            nan_log_Z_var = jnp.isnan(log_Z_var)
            nan_log_Z_mean = jnp.isnan(log_Z_mean)

            diff_log_Z_mean = jnp.abs(body_state.last_log_Z_mean - log_Z_mean)
            small_change = jnp.square(2. * diff_log_Z_mean) < log_Z_var
            decreasing_change = diff_log_Z_mean < body_state.last_diff_log_Z_mean
            stable = decreasing_change | small_change
            reset_patience = jnp.bitwise_not(stable) | (nan_log_Z_mean | nan_last_log_Z_mean | nan_log_Z_var)
            converged_step_count = jnp.where(reset_patience, jnp.asarray(0, int_type),
                                             body_state.converged_step_count + jnp.asarray(1, int_type))

            # output_log_Z = body_state.output_log_Z.at[body_state.iter_j].set(jnp.sum(sample_stats.num_live_points))
            output_log_Z = body_state.output_log_Z.at[body_state.iter_j].set(log_Z_mean)  # TODO: remove
            return CarryType(key=key, non_iid_reservoir=update_reservoir, last_log_Z_mean=log_Z_mean,
                             last_diff_log_Z_mean=diff_log_Z_mean, converged_step_count=converged_step_count,
                             state=all_state,
                             output_log_Z=output_log_Z, iter_j=body_state.iter_j + 1)

        key, improve_key = random.split(iid_state.key, 2)
        init_state = self._combine_state(state=iid_state, update_reservoir=non_iid_reservoir)
        init_evidence_calculation, _ = analyse_sample_collection(
            sample_collection=init_state.sample_collection,
            sorted_collection=True
        )
        init_log_Z_mean, _ = linear_to_log_stats(
            log_f_mean=init_evidence_calculation.log_Z_mean,
            log_f2_mean=init_evidence_calculation.log_Z2_mean
        )
        output_log_Z = jnp.full((20,), jnp.nan, float_type)
        output_log_Z = output_log_Z.at[0].set(init_log_Z_mean)
        init_body_state: CarryType = CarryType(
            key=improve_key, non_iid_reservoir=non_iid_reservoir, last_log_Z_mean=init_log_Z_mean,
            last_diff_log_Z_mean=jnp.asarray(0., float_type), converged_step_count=jnp.asarray(0, int_type),
            state=init_state, output_log_Z=output_log_Z, iter_j=jnp.asarray(1, int_type))
        output_state = while_loop(cond,
                                  body,
                                  init_body_state)

        return output_state.state, output_state.output_log_Z

    def __call__(self, state: NestedSamplerState) -> NestedSamplerState:
        """
        Adaptively refines the state until all samples are i.i.d. (according to a given stopping condition).

        Args:
            state: nested sampler state

        Returns:
            state after refinement
        """
        # Split state will give an error if this is a tracer context so don't JIT compile.
        iid_state, non_iid_reservoir = self._split_state(state)

        keys = random.split(state.key, self.num_parallel_samplers)

        non_iid_count = non_iid_reservoir.log_L.size
        remainder = int(non_iid_count) % self.num_parallel_samplers
        extra = (self.num_parallel_samplers - remainder) % self.num_parallel_samplers
        if extra > 0:
            logger.warning(
                f"Increasing non_iid_reservoir ({non_iid_count}) by {extra} to closest multiple of num_parallel_samplers.")
            non_iid_reservoir = tree_map(lambda x: jnp.concatenate([x] + [x[0:1]] * extra, axis=0), non_iid_reservoir)

        chunked_non_iid_reservoir = add_chunk_dim(non_iid_reservoir, self.num_parallel_samplers)

        parallel_ar = pmap(
            lambda key, non_iid_reservoir: self._single_improve_thread(
                key=key,
                iid_state=iid_state,
                non_iid_reservoir=non_iid_reservoir
            ),
            axis_name='i')

        chunked_updated_state, chunked_output_log_Z = parallel_ar(keys, chunked_non_iid_reservoir)

        # Is there a better way to get only one out, since they are identical on each device?
        (updated_state, output_log_Z) = tree_map(lambda x: x[0], (chunked_updated_state, chunked_output_log_Z))

        # import pylab as plt
        # plt.plot(output_log_Z)
        # plt.show()

        return updated_state



from jax import numpy as jnp
from jax.lax import scan


def msqrt(A):
    """
    Computes the matrix square-root using SVD, which is robust to poorly conditioned covariance matrices.
    Computes, M such that M @ M.T = A

    Args:
        A: [N,N] Square matrix to take square root of.

    Returns: [N,N] matrix.
    """
    U, s, Vh = jnp.linalg.svd(A)
    L = U * jnp.sqrt(s)
    return L


def squared_norm(x1, x2):
    # r2_ij = sum_k (x_ik - x_jk)^2
    #       = sum_k x_ik^2 - 2 x_jk x_ik + x_jk^2
    #       = sum_k x_ik^2 + x_jk^2 - 2 U U^T
    # r2_ij = sum_k (x_ik - y_jk)^2
    #       = sum_k x_ik^2 - 2 y_jk x_ik + y_jk^2
    #       = sum_k x_ik^2 + y_jk^2 - 2 U Y^T
    x1 = x1
    x2 = x2
    r2 = jnp.sum(jnp.square(x1), axis=1)[:, None] + jnp.sum(jnp.square(x2), axis=1)[None, :]
    r2 = r2 - 2. * (x1 @ x2.T)
    return r2




from typing import NamedTuple, Tuple

import numpy as np
import pylab as plt
from etils.array_types import IntArray, FloatArray, PRNGKey, BoolArray
from jax import numpy as jnp, vmap, random, tree_map
from jax._src.lax.control_flow import while_loop
from jax._src.scipy.special import gammaln

from jaxns.internals.log_semiring import LogSpace
from jaxns.likelihood_samplers.em_gmm import em_gmm
from jaxns.types import UType, int_type, float_type

__all__ = ['ellipsoid_clustering',
           'sample_multi_ellipsoid',
           'MultEllipsoidState']


class MultiEllipsoidParams(NamedTuple):
    mu: FloatArray  # [K, D] Ellipsoids centres
    radii: FloatArray  # [K, D] Ellsipoids radii
    rotation: FloatArray  # [K, D, D] Ellipsoids rotation matrices


class MultEllipsoidState(NamedTuple):
    params: MultiEllipsoidParams
    cluster_id: IntArray  # [N] the cluster index of each point


def log_ellipsoid_volume(radii):
    D = radii.shape[0]
    return jnp.log(2.) - jnp.log(D) + 0.5 * D * jnp.log(jnp.pi) - gammaln(0.5 * D) + jnp.sum(jnp.log(radii))


def bounding_ellipsoid(points: UType, mask: FloatArray) -> Tuple[FloatArray, FloatArray]:
    """
    Use empirical mean and covariance as approximation to bounding ellipse.

    Args:
        points: [N, D] points to fit ellipsoids to
        mask: [N] mask of which points to consider

    Returns:
        mu, cov
    """
    mu = jnp.average(points, weights=mask, axis=0)
    dx = points - mu
    cov = jnp.average(dx[:, :, None] * dx[:, None, :], weights=mask, axis=0)
    return mu, cov


def covariance_to_rotational(cov: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    (x - mu)^T inv(cov) (x - mu) = (x - mu)^T J @ J.T (x - mu)

    where J.T is composed of un-rotation and un-scaling:

    J.T = diag(1/radii) @ rotation.T <==> J = rotation @ diag(1/radii)

    Now since, cov = U @ diag(s) @ V.H we have

    J @ J.T = inv(U @ diag(s) @ V.H) = V @ diag(1/s) @ U.H

    ==> J.T = diag(1/sqrt(s)) @ U.H
    ==> radii = sqrt(s), rotation = U

    Args:
        cov:

    Returns:
        radii, rotation
    """
    u, s, vh = jnp.linalg.svd(cov)
    radii_min = jnp.finfo(s.dtype).eps
    radii = jnp.maximum(jnp.sqrt(s), radii_min)
    rotation = u
    return radii, rotation


def ellipsoid_params(points: UType, mask: FloatArray) -> Tuple[FloatArray, FloatArray, FloatArray]:
    """
    If the ellipsoid is defined by

    (x - mu)^T cov (x - mu) = 1

    where cov = L @ L.T and L = diag(1/radii) @ rotation.T

    then this returns the mu, radius and rotation matrices of the ellipsoid.

    Args:
        points: [N, D] points to fit ellipsoids to
        mask: [N] mask of which points to consider

    Returns:
        mu [D], radii [D] rotation [D,D]
    """
    # get ellipsoid mean and covariance
    mu, Sigma = bounding_ellipsoid(points=points, mask=mask)
    radii, rotation = covariance_to_rotational(Sigma)

    # Compute scale factor for radii to enclose all points.
    # for all i (points[i] - mu) @ inv(Sigma) / scale**2 @ (points[i] - mu) <= 1
    # for all i (points[i] - mu) @ (L @ L.T) @ (points[i] - mu) <= scale**2
    rho = vmap(lambda x: maha_ellipsoid(x=x, mu=mu, radii=radii, rotation=rotation))(points)

    rho_max = jnp.max(jnp.where(mask, rho, 0.))
    radii *= jnp.sqrt(rho_max)

    return mu, radii, rotation


def ellipsoid_to_circle(point: FloatArray, mu: FloatArray, radii: FloatArray, rotation: FloatArray) -> FloatArray:
    """
    Apply a linear map that would turn an ellipsoid into a sphere.
    Args:
        point: [D] point to transform
        mu: [D] center of ellipse
        radii: [D] radii of ellipse
        rotation: [D,D] rotation matrix of ellipse

    Returns:
        a transformed point of shape [D]
    """
    return jnp.diag(jnp.reciprocal(radii)) @ rotation.T @ (point - mu)


def circle_to_ellipsoid(point: FloatArray, mu: FloatArray, radii: FloatArray, rotation: FloatArray) -> FloatArray:
    """
    Apple a linear map that would turn a sphere into an ellipsoid

    Args:
        point: [D] point to transform
        mu: [D] center of ellipse
        radii: [D] radii of ellipse
        rotation: [D,D] rotation matrix of ellipse

    Returns:
        a transformed point of shape [D]
    """
    return mu + (rotation @ jnp.diag(radii) @ point)


def maha_ellipsoid(x: FloatArray, mu: FloatArray, radii: FloatArray, rotation: FloatArray) -> FloatArray:
    """
    Compute the Mahalanobis distance.

    Args:
        x: point [D]
        mu: center of ellipse [D]
        radii: radii of ellipse [D]
        rotation: rotation matrix [D, D]

    Returns:
        The Mahalanobis distance of `x` to `mu`.
    """
    u_circ = ellipsoid_to_circle(x, mu, radii, rotation)
    return u_circ @ u_circ


def point_in_ellipsoid(x: FloatArray, mu: FloatArray, radii: FloatArray, rotation: FloatArray) -> BoolArray:
    """
    Determine if a given point is inside a closed ellipse.

    Args:
        x: point [D]
        mu: center of ellipse [D]
        radii: radii of ellipse [D]
        rotation: rotation matrix [D, D]

    Returns:
        True iff x is inside the closed ellipse
    """
    return jnp.less_equal(maha_ellipsoid(x, mu, radii, rotation), jnp.asarray(1., x.dtype))


def sample_ellipsoid(key: PRNGKey, mu: FloatArray, radii: FloatArray, rotation: FloatArray,
                     unit_cube_constraint: bool = False) -> FloatArray:
    """
    Sample uniformly inside an ellipsoid. When unit_cube_constraint=True then reject points outside unit-cube.

    Args:
        key:
        mu: [D]
        radii: [D]
        rotation: [D,D]
        unit_cube_constraint: whether to restrict to the closed unit-cube.

    Returns:
        i.i.d. sample from ellipsoid of shape [D]
    """

    def _single_sample(key):
        direction_key, radii_key = random.split(key, 2)
        direction = random.normal(direction_key, shape=radii.shape)
        direction = direction / jnp.linalg.norm(direction)
        t = random.uniform(radii_key) ** (1. / radii.size)
        u_circ = direction * t
        R = rotation * radii
        u = R @ u_circ + mu
        return u

    def body(state):
        (key, _, _) = state
        key, sample_key = random.split(key, 2)
        u = _single_sample(sample_key)
        done = jnp.all((u <= 1) & (u >= 0))
        return (key, done, u)

    if unit_cube_constraint:
        (_, _, u) = while_loop(lambda s: ~s[1],
                               body,
                               (key, jnp.asarray(False), mu))
    else:
        u = _single_sample(key)
    return u


def compute_depth_ellipsoids(point: FloatArray, mu: FloatArray, radii: FloatArray, rotation: FloatArray,
                             constraint_unit_cube: bool = False) -> IntArray:
    """
    Compute overlap of ellipsoids at point. Points outside the domain are given infinite depth.

    Args:
        point: [D] point to compute depth at.
        mu: [K, D] means of ellispoids
        radii: [K, D] radii of ellipsoids
        rotation: [K, D, D] rotation matrices of ellipsoids
        constraint_unit_cube: bool, whether domain is clipped to closed unit-cube.

    Returns:
        scalar representing overlap of ellipsoids.
    """
    # in any of the ellipsoids
    contained_in = vmap(lambda mu, radii, rotation: point_in_ellipsoid(point, mu, radii, rotation))(mu, radii, rotation)
    depth = jnp.sum(contained_in)
    if constraint_unit_cube:
        # outside cube
        outside_unit_cube = jnp.any(point < 0.) | jnp.any(point > 1.)
        depth = jnp.where(outside_unit_cube, jnp.iinfo(depth.dtype).max, depth)
    return depth


def sample_multi_ellipsoid(key: PRNGKey, mu: FloatArray, radii: FloatArray, rotation: FloatArray,
                           unit_cube_constraint: bool = True) -> FloatArray:
    """
    Sample from a set of intersecting ellipsoids.
    When unit_cube_constraint=True then reject points outside the closed unit-cube.

    Args:
        key: PRNGKey
        mu: [K, D] centres of ellipses
        radii: [K, D] radii of ellipses
        rotation: [K,D,D] rotation matrices of ellipses

    Returns:
        a sample point i.i.d. sampled from union of ellipsoids, of shape [D]
    """
    # u(t) = R @ (x + t * num_options) + c
    #     u(t) == 1
    #     1-c = R@x + t * R@num_options
    #     t = ((1 - c) - R@x)/R@num_options
    K, D = radii.shape
    log_VE = vmap(log_ellipsoid_volume)(radii)
    log_p = log_VE  # - logsumexp(log_VE)

    def body(state):
        (i, _, key, done, _) = state
        key, accept_key, sample_key, select_key = random.split(key, 4)
        k = random.categorical(select_key, log_p)
        mu_k = mu[k, :]
        radii_k = radii[k, :]
        rotation_k = rotation[k, :, :]
        u_test = sample_ellipsoid(sample_key, mu_k, radii_k, rotation_k, unit_cube_constraint=False)
        depth = compute_depth_ellipsoids(u_test, mu, radii, rotation, constraint_unit_cube=unit_cube_constraint)
        done = random.uniform(accept_key) < jnp.reciprocal(depth)
        return (i + 1, k, key, done, u_test)

    _, k, _, _, u_accept = while_loop(lambda state: ~state[3],
                                      body,
                                      (jnp.array(0), jnp.array(0), key, jnp.array(False), jnp.zeros(D)))
    return k, u_accept


def log_coverage_scale(log_VE, log_VS, D):
    """
    Computes the required scaling relation such that
    V(E) = max(V(E), V(S))
    where the scaling is applied to each radius.
    Args:
        log_VE:
        log_VS:
        D:
    Returns:
    """
    return jnp.maximum(0., (log_VS - log_VE) / D)


def cluster_split(key: PRNGKey, points: FloatArray, mask: BoolArray, log_VS: FloatArray, log_VE: FloatArray,
                  kmeans_init: bool = True) \
        -> Tuple[IntArray, FloatArray, MultiEllipsoidParams, FloatArray, MultiEllipsoidParams, BoolArray]:
    """
    Splits a set of points into two ellipsoids such that the enclosed volume is as close to V(S) without being less.
    V(S) should be an estimate of the true volume contained by the points.

    Args:
        key: PRNGKey
        points: [N, D] points to split
        mask: [N] mask only those points which should be split
        log_VS: estimate of logV(S) of the set of points
        log_VE: logV(E) of the minimum volume enclosing ellipsoid of masked points
        kmeans_init: whether to use kmeans to initialise the clustering

    Returns:
        cluster indices of points (0 | 1),
        logV(S1),
        params of ellipsoid 1
        logV(S2),
        params of ellipsoid 2,
        whether split occured -- If False then above return values are undefined
    """

    init_key, volume_key = random.split(key, 2)
    N, D = points.shape
    n_S = jnp.sum(mask)
    # calculate bounding ellipsoid
    ###
    # input is essentially log_VS
    if kmeans_init:
        # do Euclidean kmean clustering
        cluster_id, (centers, covariances, log_weight), total_iters = em_gmm(key=init_key, data=points, mask=mask,
                                                                             n_components=2, n_iters=100)
    else:
        # assign to random clusters: child0 or child1
        cluster_id = random.randint(init_key, shape=(N,), minval=0, maxval=2)

    def body(state):
        (i, done, old_cluster_id, _, _, _, _, _, _, _, _, min_loss, delay) = state
        mask1 = mask & (old_cluster_id == 0)
        mask2 = mask & (old_cluster_id == 1)
        # estimate volumes of current clustering
        n1 = jnp.sum(mask1)
        n2 = jnp.sum(mask2)
        log_VS1 = log_VS + jnp.log(n1) - jnp.log(n_S)
        log_VS2 = log_VS + jnp.log(n2) - jnp.log(n_S)
        # construct E_1, E_2 and compute volumes
        mu1, radii1, rotation1 = ellipsoid_params(points=points, mask=mask1)
        log_VE1 = log_ellipsoid_volume(radii1)
        mu2, radii2, rotation2 = ellipsoid_params(points=points, mask=mask2)
        log_VE2 = log_ellipsoid_volume(radii2)
        # enlarge to at least cover V(S1) and V(S2)
        log_scale1 = log_coverage_scale(log_VE1, log_VS1, D)
        log_scale2 = log_coverage_scale(log_VE2, log_VS2, D)
        radii1 = jnp.exp(jnp.log(radii1) + log_scale1)
        radii2 = jnp.exp(jnp.log(radii2) + log_scale2)
        log_VE1 = log_VE1 + log_scale1 * D
        log_VE2 = log_VE2 + log_scale2 * D
        # compute reassignment metrics
        maha1 = vmap(lambda point: maha_ellipsoid(point, mu=mu1, radii=radii1, rotation=rotation1))(points)
        maha2 = vmap(lambda point: maha_ellipsoid(point, mu=mu2, radii=radii2, rotation=rotation2))(points)
        log_h1 = log_VE1 - log_VS1 + jnp.log(maha1)
        log_h2 = log_VE2 - log_VS2 + jnp.log(maha2)
        # reassign
        delta_F = LogSpace(log_h1) - LogSpace(log_h2)
        reassign_idx = jnp.argmax(delta_F.abs().log_abs_val)
        new_id = jnp.asarray(delta_F[reassign_idx] > LogSpace(-jnp.inf), int_type)
        new_cluster_id = cluster_id.at[reassign_idx].set(new_id)

        # new_cluster_k = jnp.where(log_h1 < log_h2, 0, 1)
        log_V_sum = jnp.logaddexp(log_VE1, log_VE2)
        new_loss = log_V_sum - log_VS
        loss_decreased = new_loss < min_loss
        delay = jnp.where(loss_decreased, 0, delay + 1)
        min_loss = jnp.where(loss_decreased, new_loss, min_loss)
        ###
        # i / delay / loss_decreased / new_loss / min_loss
        # 0 / 0 / True / a / a
        # 1 / 1 / False / b / a
        # 2 / 2 / False / a / a
        # 3 / 3 / False / b / a
        # 4 / 4 / False / a / a
        done = jnp.all(new_cluster_id == old_cluster_id) \
               | (delay >= 10) \
               | (n1 < D + 1) \
               | (n2 < D + 1) \
               | jnp.isnan(log_V_sum)
        return (
            i + jnp.asarray(1, int_type), done, new_cluster_id, log_VS1, mu1, radii1, rotation1, log_VS2, mu2, radii2,
            rotation2,
            min_loss, delay)

    done = (jnp.sum(mask) < 2 * (D + 1))
    (i, _, cluster_id, log_VS1, mu1, radii1, rotation1, log_VS2, mu2, radii2, rotation2, min_loss, delay) = \
        while_loop(lambda state: ~state[1],
                   body,
                   (jnp.array(0), done, cluster_id,
                    jnp.array(-jnp.inf), jnp.zeros(D), jnp.zeros(D), jnp.eye(D),
                    jnp.array(-jnp.inf), jnp.zeros(D), jnp.zeros(D), jnp.eye(D),
                    jnp.asarray(jnp.inf), jnp.asarray(0, int_type)))
    mask1 = mask & (cluster_id == 0)
    mask2 = mask & (cluster_id == 1)
    log_V_sum = jnp.logaddexp(log_ellipsoid_volume(radii1), log_ellipsoid_volume(radii2))

    # V(A v B) = V(A) + V(B) - V(A ^ B) <= V(A) + V(B)
    # V(A v B) / (V(A) + V(B)) = 1 - V(A ^ B)/(V(A) + V(B))
    # (1 - V(A v B) / (V(A) + V(B))) = V(A ^ B)/(V(A) + V(B))
    # V(A ^ B) / V(A v B) = (1 - V(A v B) / (V(A) + V(B))) (V(A) + V(B)) / V(A v B)
    # intersection_ratio = jnp.log(1. - jnp.exp(log_V_union - log_V_sum)) + log_V_sum - log_V_union
    # no_intersection = jnp.exp(intersection_ratio) < 0.05
    no_intersection = (log_V_sum < log_VE)

    do_split = (no_intersection | (log_VE > log_VS + jnp.log(2.))) \
               & (~jnp.any(jnp.isnan(radii1))) \
               & (~jnp.any(jnp.isnan(radii2))) \
               & (jnp.sum(mask1) >= (D + 1)) \
               & (jnp.sum(mask2) >= (D + 1))

    # do_split = (log_VE > log_VS + jnp.log(2.)) \
    #            | ((~jnp.any(jnp.isnan(radii1)))
    #               & (~jnp.any(jnp.isnan(radii2)))
    #               & (jnp.sum(mask1) >= (D + 1))
    #               & (jnp.sum(mask2) >= (D + 1)))
    # & (cond1 < 50.) \
    # & (cond2 < 50.)

    params1 = MultiEllipsoidParams(mu=mu1,
                                   radii=radii1,
                                   rotation=rotation1)

    params2 = MultiEllipsoidParams(mu=mu2, radii=radii2, rotation=rotation2)

    return cluster_id, log_VS1, params1, log_VS2, params2, do_split


def plot_ellipses(params: MultiEllipsoidParams):
    theta = jnp.linspace(0., 2 * jnp.pi, 100)
    circle = jnp.stack([jnp.cos(theta), jnp.sin(theta)], axis=1)
    for mu, radii, rotation in zip(params.mu, params.radii, params.rotation):
        ellipse = vmap(lambda point: circle_to_ellipsoid(point, mu, radii, rotation))(circle)
        plt.plot(ellipse[:, 0], ellipse[:, 1], c=np.random.uniform(size=3))
    plt.show()


def ellipsoid_clustering(key: PRNGKey, points: FloatArray, log_VS: FloatArray,
                         max_num_ellipsoids: int) -> MultEllipsoidState:
    """
    Partition live_points into 2^depth ellipsoids in depth-first order.

    Args:
        key:PRNGKey
        points: [N, D] points to partition
        log_VS: expected true volume of points
        max_num_ellipsoids: the maximum number of ellipsoids

    Returns:
        params of multi-ellipsoids and cluster id of points
    """

    N, D = points.shape
    if max_num_ellipsoids < 1:
        raise ValueError(f"max_num_ellipsoids should be >= 1, got {max_num_ellipsoids}.")
    K = max_num_ellipsoids

    cluster_id = jnp.zeros(N, dtype=int_type)

    params = MultiEllipsoidParams(
        mu=jnp.zeros((K, D), float_type),
        radii=jnp.zeros((K, D), float_type),
        rotation=jnp.zeros((K, D, D), float_type)
    )

    mu, radii, rotation = ellipsoid_params(points=points, mask=jnp.ones(N, jnp.bool_))

    params = params._replace(
        mu=params.mu.at[0].set(mu),
        radii=params.radii.at[0].set(radii),
        rotation=params.rotation.at[0].set(rotation)
    )

    state = MultEllipsoidState(cluster_id=cluster_id,
                               params=params)

    log_VS_subclusters = jnp.array([log_VS] + [-jnp.inf] * (K - 1))
    done_splitting = jnp.asarray([False] + [True] * (K - 1), jnp.bool_)
    split_depth = jnp.zeros([K], int_type)

    CarryType = Tuple[PRNGKey, IntArray, MultEllipsoidState, BoolArray, IntArray, FloatArray]

    def body(body_state: CarryType) -> CarryType:
        (key, next_k, state, done_splitting, split_depth, log_VS_subclusters) = body_state

        key, split_key = random.split(key, 2)
        # bread first selection
        select_split = jnp.argmin(jnp.where(done_splitting, jnp.iinfo(split_depth.dtype).max, split_depth))

        mask = state.cluster_id == select_split
        log_VS_subcluster = log_VS_subclusters[select_split]
        log_VE_parent = log_ellipsoid_volume(state.params.radii[select_split])

        unsorted_cluster_id, log_VS1, params1, log_VS2, params2, do_split = cluster_split(
            split_key, points, mask, log_VS_subcluster, log_VE_parent, kmeans_init=True)

        params = tree_map(lambda x, y: jnp.where(do_split, x.at[select_split].set(y), x),
                          state.params,
                          params1)
        params = tree_map(lambda x, y: jnp.where(do_split, x.at[next_k].set(y), x),
                          params,
                          params2)

        # update those points where mask is true, unsorted id is 1, and do_split is true
        cluster_id = jnp.where(do_split & (unsorted_cluster_id == 1) & mask, next_k, state.cluster_id)

        state = state._replace(params=params, cluster_id=cluster_id)

        done_splitting = jnp.where(do_split,
                                   done_splitting.at[next_k].set(False),
                                   done_splitting.at[select_split].set(True))

        new_depth = split_depth[select_split] + jnp.asarray(1, int_type)
        split_depth = jnp.where(do_split, split_depth.at[select_split].set(new_depth), split_depth)
        split_depth = jnp.where(do_split, split_depth.at[next_k].set(new_depth), split_depth)

        log_VS_subclusters = jnp.where(do_split, log_VS_subclusters.at[select_split].set(log_VS1), log_VS_subclusters)
        log_VS_subclusters = jnp.where(do_split, log_VS_subclusters.at[next_k].set(log_VS2), log_VS_subclusters)

        # if no split we replace child0 with parent and child1 gets zero-size ellipsoid that has no members.

        next_k = next_k + jnp.asarray(1, int_type)

        return (key, next_k, state, done_splitting, split_depth, log_VS_subclusters)

    def cond(body_state: CarryType) -> BoolArray:
        (key, next_k, state, done_splitting, split_depth, log_VS_subclusters) = body_state
        done = jnp.all(done_splitting) | (next_k == K)
        return jnp.bitwise_not(done)

    init_body_state = (key, jnp.asarray(1, int_type), state, done_splitting, split_depth, log_VS_subclusters)
    (key, next_k, state, done_splitting, split_depth, log_VS_subclusters) = \
        while_loop(cond,
                   body,
                   init_body_state)

    return state



import numpy as np
from jax import numpy as jnp, random

from jaxns.internals.stats import linear_to_log_stats, density_estimation


def test_linear_to_log_stats():
    Z = jnp.exp(random.normal(key=random.PRNGKey(42), shape=(1000000,)))
    Z_mu = jnp.mean(Z)
    Z_var = jnp.var(Z)
    log_mu1, log_var1 = linear_to_log_stats(jnp.log(Z_mu), log_f_var=jnp.log(Z_var))
    log_mu2, log_var2 = linear_to_log_stats(jnp.log(Z_mu), log_f2_mean=jnp.log(Z_var + Z_mu ** 2))
    assert jnp.isclose(log_mu1, 0., atol=1e-2)
    assert jnp.isclose(log_var1, 1., atol=1e-2)
    assert jnp.isclose(log_mu1,  log_mu2, atol=1e-4)
    assert jnp.isclose(log_var1, log_var2, atol=1e-4)


def test_density_estimation():
    np.random.seed(42)
    x = jnp.asarray(np.random.standard_gamma(1., 100))[:, None]
    xstar = jnp.linspace(0., 20., 1000)[:, None]
    assert density_estimation(xstar, x).size == 1000
    #
    # import pylab as plt
    #
    # plt.plot(xstar, density_estimation(xstar, x))
    # plt.hist(np.random.standard_gamma(1., 10000), bins=np.linspace(0, 20, 100), density=True, alpha=0.5)
    # plt.hist(x[:, 0], bins=np.linspace(0., 20, 100), density=True, alpha=0.5)
    # plt.show()



import logging
from typing import Tuple, NamedTuple

from etils.array_types import PRNGKey, FloatArray, BoolArray, IntArray
from jax import tree_map, numpy as jnp, random, core
from jax._src.lax.control_flow import scan, while_loop

from jaxns.model import Model
from jaxns.random import resample_indicies
from jaxns.types import Reservoir, SampleCollection, LivePoints, NestedSamplerState, float_type, \
    int_type, Sample
from jaxns.utils import sort_samples

logger = logging.getLogger('jaxns')

__all__ = ['get_live_points_from_samples',
           'get_uniform_init_live_points']


def init_sample_collection(size: int, model: Model) -> SampleCollection:
    """
    Return an initial sample collection, that will be incremented by the sampler.

    Args:
        size: the size of the sample collection.

    Returns:
        sample collection
    """

    def _repeat(a):
        return jnp.repeat(a[None], repeats=size, axis=0, total_repeat_length=size)

    reservoir = Reservoir(
        point_U=_repeat(model.U_placeholder),
        log_L_constraint=jnp.full((size,), jnp.inf, dtype=float_type),
        log_L=jnp.full((size,), jnp.inf, dtype=float_type),
        num_likelihood_evaluations=jnp.full((size,), 0, dtype=int_type),
        num_slices=jnp.full((size,), 0, dtype=int_type),
        iid=jnp.full((size,), False, dtype=jnp.bool_)
    )

    sample_collection = SampleCollection(
        sample_idx=jnp.zeros((), int_type),
        reservoir=reservoir
    )
    return sample_collection


def _single_uniform_sample(key: PRNGKey, model: Model) -> Sample:
    """
    Gets a single sample strictly within -inf bound (the entire prior), i.e. all returned samples will have non-zero
    likeihood.

    Args:
        key: PRNGKey
        model: the model to use.

    Returns:
        a sample
    """

    class CarryState(NamedTuple):
        done: BoolArray
        key: PRNGKey
        U: FloatArray
        log_L: FloatArray
        num_likelihood_evals: IntArray

    def body(carry_state: CarryState):
        key, sample_key = random.split(carry_state.key, 2)
        log_L = model.forward(U=carry_state.U)
        num_likelihood_evals = carry_state.num_likelihood_evals + jnp.asarray(1, int_type)
        done = log_L > -jnp.inf
        U = jnp.where(done, carry_state.U, model.sample_U(key=sample_key))
        return CarryState(done=done, key=key, U=U, log_L=log_L, num_likelihood_evals=num_likelihood_evals)

    key, sample_key = random.split(key, 2)
    init_carry_state = CarryState(done=jnp.asarray(False),
                                  key=key,
                                  U=model.sample_U(key=sample_key),
                                  log_L=-jnp.inf,
                                  num_likelihood_evals=jnp.asarray(0, int_type))

    carry_state = while_loop(lambda s: jnp.bitwise_not(s.done), body, init_carry_state)

    sample = Sample(point_U=carry_state.U,
                    log_L_constraint=-jnp.inf,
                    log_L=carry_state.log_L,
                    num_likelihood_evaluations=carry_state.num_likelihood_evals,
                    num_slices=jnp.asarray(0, int_type),
                    iid=jnp.asarray(True, jnp.bool_))
    return sample


def get_uniform_init_live_points(key: PRNGKey, num_live_points: int, model: Model) -> LivePoints:
    """
    Get initial live points from uniformly sampling the entire prior.

    Args:
        key: PRNGKey
        num_live_points: the number of live points
        model: the model

    Returns:
        live points
    """
    _, samples = scan(lambda s, key: (s, _single_uniform_sample(key=key, model=model)),
                      (),
                      random.split(key, num_live_points))
    reservoir = Reservoir(*samples)
    return LivePoints(reservoir=reservoir)


def sort_sample_collection(sample_collection: SampleCollection) -> SampleCollection:
    """
    Sort a sample collection lexigraphically.

    Args:
        sample_collection: sample collections

    Returns:
        sample collection sorted
    """
    idx_sort = jnp.lexsort((sample_collection.reservoir.log_L_constraint,
                            sample_collection.reservoir.log_L))
    sample_collection = sample_collection._replace(
        reservoir=tree_map(lambda x: x[idx_sort], sample_collection.reservoir)
    )
    return sample_collection


def get_live_points_from_samples(state: NestedSamplerState, log_L_constraint: FloatArray, num_live_points: int,
                                 sorted_collection: bool = True) \
        -> Tuple[NestedSamplerState, LivePoints]:
    """
    Extract live points from the samples already collected.

    Args:
        state: the current state
        log_L_constraint: the contour to sample above
        num_live_points: the number of unique live points to sample
        sorted_collection: whether the sample collection is already sorted

    Returns:
        a new state, and live points
    """
    if isinstance(state.sample_collection.reservoir.iid, core.Tracer):
        raise RuntimeError("Tracer detected, but expected imperative context.")
    ###
    # We select from samples where sample.log_L_constraint <= log_L_constraint
    # 1. sort samples
    # 2. find the largest index, s, where sample.log_L_constraint[s] <= log_L_constraint
    # 3. select without replacement `num_live_points` indices from 0..s (inclusive)
    # 4. replace those indices with an empty sample

    sample_collection = state.sample_collection
    if not sorted_collection:
        sample_collection = sort_sample_collection(sample_collection)

    # find the largest index, s, where sample.log_L_constraint[s] <= log_L_constraint
    idx_supremum = jnp.searchsorted(sample_collection.reservoir.log_L_constraint, log_L_constraint, side='right') - 1
    idx_supremum = int(idx_supremum)
    key, sample_key = random.split(state.key)
    take_indices = resample_indicies(key=sample_key,
                                     log_weights=None,
                                     S=num_live_points,
                                     replace=False,
                                     num_total=idx_supremum + 1)
    reservoir = tree_map(lambda x: x[take_indices], sample_collection.reservoir)
    live_points = LivePoints(reservoir=reservoir)

    # replace the taken indicies
    empty_sample: Reservoir = tree_map(lambda x: jnp.zeros_like(x[-1]), sample_collection.reservoir)
    empty_sample = empty_sample._replace(log_L_constraint=jnp.inf, log_L=jnp.inf)

    sample_collection = sample_collection._replace(
        reservoir=tree_map(lambda x, y: x.at[take_indices].set(y), sample_collection.reservoir, empty_sample),
        sample_idx=sample_collection.sample_idx - num_live_points
    )
    # sort the return
    sample_collection = sort_samples(sample_collection=sample_collection)
    state = state._replace(key=key, sample_collection=sample_collection)
    return state, live_points



import logging
from functools import wraps
from typing import Callable, Optional

logger = logging.getLogger('jaxns')


def deprecated(other: Optional[Callable]) -> Callable:
    def decorator(f: Callable):
        @wraps
        def g(*args, **kwargs):
            if other is not None:
                logger.warning(f"{f.__name__} is deprecated, use `{other.__name__}`.")
            else:
                logger.warning(f"{f.__name__} is deprecated.")
            return f(*args, **kwargs)

        if other is not None:
            g.__doc__ = (f"Deprecated. Prefer {other.__name__}.\n\n{g.__doc__}")
        else:
            g.__doc__ = (f"Deprecated.\n\n{g.__doc__}")

        return g

    return decorator
